/*
 * atapi.c - Copyright (C) 1999 Jay A. Key
 *
 * Functions for the CDR_ATAPI1, CDR_ATAPI2 class of read/init/deinit functions
 *
 **********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 **********************************************************************
 *
 */

#define _AKRIP32_

#include <windows.h>
#include <stdio.h>
#include "myaspi32.h"
#include "scsidefs.h"
#include "aspilib.h"
#include "akrip32.h"

#include "NVExtension.h"


extern CDHANDLEREC *cdHandles;
extern HANDLE *cdMutexes;
extern CRITICAL_SECTION getHandle;
extern int alErrCode;
extern BYTE alAspiErr;
extern int *nextHandle;

extern DWORD (*pfnSendASPI32Command)(LPSRB);

DWORD testUnitReady_ATAPI( HCDROM hCD );



/***************************************************************************
 * readCDAudioLBA_ATAPI
 ***************************************************************************/
DWORD readCDAudioLBA_ATAPI( HCDROM hCD, LPTRACKBUF t )
{
  DWORD dwStatus;
  DWORD dwStatusLock = SS_ERR;
  DWORD dwStatusEnable = SS_ERR;
  int idx = (int)hCD - 1;

#ifdef _DEBUG_BUILD
	  dbprintf("!!!Akrip32: Start Read CD-DA ATAPI");
#endif
  if ( (idx<0) || (idx>=MAXCDHAND) || !cdHandles[idx].used )
    {
#ifdef _DEBUG_BUILD
	  dbprintf("!!!Akrip32: Invalid CD Handle idx = %d : used = %d",idx, (idx >= 0 && idx <= MAXCDHAND) ? cdHandles[idx].used : -1);
#endif
      alErrCode = ALERR_INVHANDLE;
      return SS_ERR;
    }


  if ( t->numFrames * 2352 > t->maxLen )
    {
#ifdef _DEBUG_BUILD
	  dbprintf("!!!Akrip32: Buffer Too Small. frames = %d, maxlen = %d", t->numFrames, t->maxLen);
#endif
      alErrCode = ALERR_BUFTOOSMALL;
      return SS_ERR;
    }

  dwStatus = WaitForSingleObject( cdMutexes[idx], TIMEOUT );
  if ( dwStatus != WAIT_OBJECT_0 )
    {
#ifdef _DEBUG_BUILD
	  if (dwStatus == WAIT_TIMEOUT)
	  {
		  dbprintf("!!!Akrip32: Mutex TimeOut");
	  }
	  else
	  {
		  dbprintf("!!!Akrip32: Mutex Error");
	  }
#endif
      alErrCode = ALERR_LOCK;
      return SS_ERR;
    }

#ifdef _DEBUG_BUILD
  dbprintf( "akrip32: readCDAudioLBA_ATAPI: (%d:%d:%d) %06X:%02X", 
	    cdHandles[idx].ha, cdHandles[idx].tgt, cdHandles[idx].lun,
	    t->startFrame, t->numFrames );
#endif

   if (testUnitReady_ATAPI( hCD ) != SS_COMP)
   {
#ifdef _DEBUG_BUILD
	  dbprintf("!!!Akrip32: Unit Not Ready");
#endif

      alErrCode = ALERR_ASPI;
      ReleaseMutex( cdMutexes[idx] );
	   return SS_ERR;
   }

/*  {
	  DWORD dwSpeed = 32 * 0xB0;
	  QueryCDParms(hCD, CDP_MAXSPEED, &dwSpeed);
	  if (dwSpeed)
	  {
		  ModifyCDParms(hCD, CDP_CURRSPEED, dwSpeed ? dwSpeed : 32 * 0xB0);
	  }
  }
*/
   if (cdHandles[idx].fLockDrive)
   {
#ifdef _DEBUG_BUILD
		dbprintf( "!!!akrip32: LockDrive");
#endif
	   dwStatusLock = PreventMediaRemoval(TRUE, hCD);
	   if (dwStatusLock != SS_COMP)
	   {
#ifdef _DEBUG_BUILD
		   dbprintf("!!!akrip32: Failed LockDrive");
#endif
		   cdHandles[idx].fLockDrive = FALSE;
	   }
   }
   if (cdHandles[idx].fSetSector)
   {
#ifdef _DEBUG_BUILD
	   dbprintf("!!!akrip32: SetSectorMode");
#endif
	   dwStatusEnable = EnableCdda(TRUE, hCD);
	   if (dwStatusEnable != SS_COMP)
	   {
#ifdef _DEBUG_BUILD
		   dbprintf("!!!akrip32: Fialed SetSectorMode");
#endif
		   cdHandles[idx].fSetSector = FALSE;
	   }
   }

  {
	  BYTE cmd[12];
	  memset(cmd, 0x00, sizeof(cmd));

	  cmd[0] = 0xBE;
#if 0
	  cmd[1] = 0x04;
#else
	  cmd[1] = (cdHandles[idx].lun << 5) | ((cdHandles[idx].readType == CDR_ATAPI1 || cdHandles[idx].readType == CDR_ATAPI2) ? 0x00 : 0x04);
#endif
	  cmd[2] = (BYTE)((t->startFrame >> 24) & 0xFF);
	  cmd[3] = (BYTE)((t->startFrame >> 16) & 0xFF);
	  cmd[4] = (BYTE)((t->startFrame >> 8) & 0xFF);
	  cmd[5] = (BYTE)(t->startFrame & 0xFF);
	  cmd[7] = (BYTE)((t->numFrames >> 8) & 0xFF);
	  cmd[8] = (BYTE)(t->numFrames & 0xFF);
#if 0
	  cmd[9] = (cdHandles[idx].readType == CDR_ATAPI1) ? 0x10 : 0xF0;
#else
	  cmd[9] = (cdHandles[idx].readType == CDR_ATAPI1 || cdHandles[idx].readType == CDR_ATAPI3) ? 0x10 : 0xF8;
#endif

	  dwStatus = IssueSCSICommand(SRB_DIR_IN, cmd, sizeof(cmd), &(t->buf[0]), t->maxLen, hCD);
  }

  if (dwStatusEnable == SS_COMP)
  {
#ifdef _DEBUG_BUILD
	  dbprintf("!!!akrip32: Restore SectorMode");
#endif
	  EnableCdda(FALSE, hCD);
  }
  if (dwStatusLock == SS_COMP)
  {
#ifdef _DEBUG_BUILD
	  dbprintf("!!!akrip32: UnLockDrive");
#endif
	  PreventMediaRemoval(FALSE, hCD);
  }

  if ( dwStatus != SS_COMP )
    {
#if 0
#ifdef _DEBUG_BUILD
      BYTE *p;
      dbprintf( "akrip32: readCDAudioLBA_ATAPI: ERROR! 0x%08X\n", s.SRB_Status );
      dbprintf( "akrip32:  haStat == %d (0x%04X), tgtStat == %d (0x%04X)",
	       s.SRB_HaStat, s.SRB_HaStat, s.SRB_TargStat, s.SRB_TargStat );
      p = s.SenseArea;
      dbprintf( "  %02X %02X %02X %02X %02X %02X %02X %02X",
	       p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7] );
      p = &(s.SenseArea[8]);
      dbprintf( "  %02X %02X %02X %02X %02X %02X %02X %02X",
	       p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7] );
#endif
#endif
      alErrCode = ALERR_ASPI;
      alAspiErr = (BYTE)dwStatus;
      ReleaseMutex( cdMutexes[idx] );
      return SS_ERR;
    }

#ifdef _DEBUG_BUILD
	  dbprintf("!!!Akrip32: Read CD-DA Success");
#endif
  t->len = t->numFrames * 2352;
  t->startOffset = 0;
  t->status = dwStatus;

  ReleaseMutex( cdMutexes[idx] );
  return dwStatus;
}

DWORD testUnitReady_ATAPI( HCDROM hCD )
{
  DWORD dwStatus;
  int idx = (int)hCD - 1;
  
  if ( (idx<0) || (idx>=MAXCDHAND) )
    {
      alErrCode = ALERR_INVHANDLE;
      return SS_ERR;
    }

  {
	  BYTE cmd[12];
	  memset(cmd, 0x00, sizeof(cmd));

	  dwStatus = IssueSCSICommand(SRB_DIR_IN, cmd, sizeof(cmd), NULL, 0, hCD);
  }
  if ( dwStatus != SS_COMP )
    {
#ifdef _DEBUG_BUILD
#endif
      alErrCode = ALERR_ASPI;
      alAspiErr = (BYTE)dwStatus;
      return SS_ERR;
    }

#ifdef _DEBUG_BUILD
  dbprintf( "akrip32: testUnitReady_ATAPI returning %d", dwStatus );
#endif
    
  return dwStatus;  
}
