#include "ReplayGain.h"
#include "GainAnalysis.h"

#define READ_SAMPLES 1152

NVRESULT __stdcall AnalyzeReplayGain(INVSTREAM invst, REPLAYGAININFO& ReplayGain, VSSENCPROGRESS VSSEncProgress, LPCVOID lpvProgressParam, DWORD dwBytesPerSample)
{
	NVRESULT nvres;
	IVSSMANAGER ivsm = NULL;
	INVAUDIOFILE invaf = NULL;

	ClearReplayGainInfo(&ReplayGain);
	try
	{
		VSSMANAGERCREATE vsmc;

		nvres = VCreateInstance(VSSMANAGER_FOURCC, VSSMANAGER_FOURCC, (LPVOID *) &ivsm);
		if (nvres)
		{
			throw nvres;
		}

		SetVSSManagerCreate(&vsmc, 0);
		nvres = ivsm->Create(&vsmc);
		if (nvres)
		{
			throw nvres;
		}

		DWORD dwFourCC;
		nvres = ivsm->DecGetFileTypeFourCCByStream(invst, &dwFourCC);

		if (nvres == NV_OK && dwFourCC)
		{
			nvres = ivsm->DecCreateInstance(dwFourCC, NVAUDIOFILE_FOURCC, (LPVOID *) &invaf);

			if (nvres)
			{
				throw nvres;
			}
		}
		else
		{
			throw VAFERR_UNSUPPORTED;
		}

		NVAUDIOFILECREATE nvafc;
		//  NVAUDIOFILECREATE 
		nvafc.dwFourCC = NVAUDIOFILE_FOURCC;
		nvafc.dwObjSize = sizeof(NVAUDIOFILECREATE);
		nvafc.dwFlag = NVAUDIOFILE_OPENSTREAM;
		nvafc.invst = invst;
		nvafc.lBitsPerSample = -64;

		//  INVAUDIOFILE 
		nvres = invaf->Create(&nvafc);
		invst = NULL;
		if (nvres)
		{
			throw nvres;
		}

		NVFPCMCREATE nvfpc;
		INVAUDIOFILE invfpcm = NULL;

		SetNVFPCMCreate(&nvfpc, 0, invaf, TRUE, 64);
		nvres = VCreateInstance(NVFPCM_FOURCC, NVAUDIOFILE_FOURCC, (LPVOID *) &invfpcm);
		if (nvres)
		{
			throw nvres;
		}
		nvres = invfpcm->Create(&nvfpc);
		if (nvres)
		{
			invfpcm->Release();
			throw nvres;
		}

		invaf = invfpcm;

		{
			PCMWAVEFORMAT pcmwf;
			invaf->GetPCMWaveFormat(&pcmwf);

			switch (pcmwf.wf.nSamplesPerSec)
			{
			case 48000:
			case 44100:
			case 32000:
			case 24000:
			case 22050:
			case 16000:
			case 12000:
			case 11025:
			case 8000:
				break;
			default:
				{
					NVSRCCREATE nvsc;
					INVAUDIOFILE invsrc = NULL;

					SetNVSRCCreate(&nvsc, 0, invaf, -64, 48000);
					nvres = VCreateInstance(NVSRC_FOURCC, NVAUDIOFILE_FOURCC, (LPVOID *) &invsrc);
					if (nvres)
					{
						throw nvres;
					}
					nvres = invsrc->Create(&nvsc);
					if (nvres)
					{
						invsrc->Release();
						throw nvres;
					}

					invaf = invsrc;
				}
			}
		}

		{
			PCMWAVEFORMAT pcmwf;
			invaf->GetPCMWaveFormat(&pcmwf);

			ReplayGainAnalysis analyzer;
			if (INIT_GAIN_ANALYSIS_OK != analyzer.InitGainAnalysis(pcmwf.wf.nSamplesPerSec))
			{
				throw NVERR_UNDEFINED;
			}

			int i, j;
			double *input = new double [READ_SAMPLES * pcmwf.wf.nChannels];
			double **output = new double *[pcmwf.wf.nChannels];
			double peak = 0.0;

			QWORD qwAnalyzeTotal = 0;

			for (i = 0; i < pcmwf.wf.nChannels; i++)
			{
				output[i] = new double [READ_SAMPLES];
			}
			DWORD dwDestBytesPerSample = sizeof(double) * pcmwf.wf.nChannels;

			while ((VSSEncProgress ? VSSEncProgress(lpvProgressParam, qwAnalyzeTotal * dwBytesPerSample / dwDestBytesPerSample) : NV_OK) == NV_OK)
			{
				LONG lResultSize = 0;
				invaf->DecodeToPCMData(input, READ_SAMPLES * sizeof(double) * pcmwf.wf.nChannels, &lResultSize);
				qwAnalyzeTotal += lResultSize;
				lResultSize /= (sizeof(double) * pcmwf.wf.nChannels);
				if (lResultSize <= 0)
				{
					break;
				}
				double *in = input;
				for (i = 0; i < lResultSize; i++)
				{
					for (j = 0; j < pcmwf.wf.nChannels; j++)
					{
						double t = *(in++);
						output[j][i] = t * 32768.0;
						if (peak < fabs(t))
						{
							peak = fabs(t);
						}
					}
				}
				analyzer.AnalyzeSamples((const Float_t **)output, lResultSize, pcmwf.wf.nChannels);
			}

			ReplayGain.dwFlag = REPLAYGAIN_TRACK_GAIN | REPLAYGAIN_TRACK_PEAK;
			ReplayGain.dTrackGain = analyzer.GetTitleGain();
			ReplayGain.dTrackPeak = peak;

			for (i = 0; i < pcmwf.wf.nChannels; i++)
			{
				delete [] output[i];
			}
			delete [] output;
			delete [] input;
		}

		invaf->Release();
		ivsm->Release();
	}
	catch (NVRESULT nvres)
	{
		if (invaf)
		{
			invaf->Release();
		}

		if (ivsm)
		{
			ivsm->Release();
		}

		return nvres;
	}

	return NV_OK;
}

NVRESULT __stdcall AnalyzeReplayGainFromNVAudioFile(INVAUDIOFILE invaf, REPLAYGAININFO& ReplayGain, VSSENCPROGRESS VSSEncProgress, LPCVOID lpvProgressParam)
{
	NVRESULT nvres;
	INVAUDIOFILE invafOrg = NULL;

	ClearReplayGainInfo(&ReplayGain);
	try
	{
		invaf->AddRef();
		invafOrg = invaf;
		invaf->Rewind();

		DWORD dwBytesPerSample;
		{
			PCMWAVEFORMAT pcmwf;
			invaf->GetPCMWaveFormat(&pcmwf);

			dwBytesPerSample = pcmwf.wBitsPerSample * pcmwf.wf.nChannels / 8;
		}
		NVFPCMCREATE nvfpc;
		INVAUDIOFILE invfpcm = NULL;

		SetNVFPCMCreate(&nvfpc, 0, invaf, TRUE, 64);
		nvres = VCreateInstance(NVFPCM_FOURCC, NVAUDIOFILE_FOURCC, (LPVOID *) &invfpcm);
		if (nvres)
		{
			throw nvres;
		}
		nvres = invfpcm->Create(&nvfpc);
		if (nvres)
		{
			invfpcm->Release();
			throw nvres;
		}

		invaf = invfpcm;

		{
			PCMWAVEFORMAT pcmwf;
			invaf->GetPCMWaveFormat(&pcmwf);

			switch (pcmwf.wf.nSamplesPerSec)
			{
			case 48000:
			case 44100:
			case 32000:
			case 24000:
			case 22050:
			case 16000:
			case 12000:
			case 11025:
			case 8000:
				break;
			default:
				{
					NVSRCCREATE nvsc;
					INVAUDIOFILE invsrc = NULL;

					SetNVSRCCreate(&nvsc, 0, invaf, -64, 48000);
					nvres = VCreateInstance(NVSRC_FOURCC, NVAUDIOFILE_FOURCC, (LPVOID *) &invsrc);
					if (nvres)
					{
						throw nvres;
					}
					nvres = invsrc->Create(&nvsc);
					if (nvres)
					{
						invsrc->Release();
						throw nvres;
					}

					invaf = invsrc;
				}
			}
		}

		{
			PCMWAVEFORMAT pcmwf;
			invaf->GetPCMWaveFormat(&pcmwf);

			ReplayGainAnalysis analyzer;
			if (INIT_GAIN_ANALYSIS_OK != analyzer.InitGainAnalysis(pcmwf.wf.nSamplesPerSec))
			{
				throw NVERR_UNDEFINED;
			}

			int i, j;
			double *input = new double [READ_SAMPLES * pcmwf.wf.nChannels];
			double **output = new double *[pcmwf.wf.nChannels];
			double peak = 0.0;

			QWORD qwAnalyzeTotal = 0;

			for (i = 0; i < pcmwf.wf.nChannels; i++)
			{
				output[i] = new double [READ_SAMPLES];
			}
			DWORD dwDestBytesPerSample = sizeof(double) * pcmwf.wf.nChannels;

			while ((VSSEncProgress ? VSSEncProgress(lpvProgressParam, qwAnalyzeTotal * dwBytesPerSample / dwDestBytesPerSample) : NV_OK) == NV_OK)
			{
				LONG lResultSize = 0;
				invaf->DecodeToPCMData(input, READ_SAMPLES * sizeof(double) * pcmwf.wf.nChannels, &lResultSize);
				qwAnalyzeTotal += lResultSize;
				lResultSize /= (sizeof(double) * pcmwf.wf.nChannels);
				if (lResultSize <= 0)
				{
					break;
				}
				double *in = input;
				for (i = 0; i < lResultSize; i++)
				{
					for (j = 0; j < pcmwf.wf.nChannels; j++)
					{
						double t = *(in++);
						output[j][i] = t * 32768.0;
						if (peak < fabs(t))
						{
							peak = fabs(t);
						}
					}
				}
				analyzer.AnalyzeSamples((const Float_t **)output, lResultSize, pcmwf.wf.nChannels);
			}

			ReplayGain.dwFlag = REPLAYGAIN_TRACK_GAIN | REPLAYGAIN_TRACK_PEAK;
			ReplayGain.dTrackGain = analyzer.GetTitleGain();
			ReplayGain.dTrackPeak = peak;

			for (i = 0; i < pcmwf.wf.nChannels; i++)
			{
				delete [] output[i];
			}
			delete [] output;
			delete [] input;
		}

		invaf->Release();
	}
	catch (NVRESULT nvres)
	{
		if (invaf)
		{
			invaf->Release();
		}

		return nvres;
	}

	invafOrg->Rewind();

	return NV_OK;
}
