#include "..\\uLilith\\Plugin\\GenericPlugin.h"
#include <shlwapi.h>
#include "Resource.h"

#include <string>
#include <xstring>


#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "gdiplus.lib")


#define TARGET_FRAMEBUFFER TEXT("AlbumJacket")
#define EXTENSIONS_COUNT 4
#define JACKET_FILENAME_COUNT 14
#define STR_LEN 1024
#define PROP_ALBUMJACKETDATA TEXT("Prop::AlbumJacketData")

#ifndef _tstring
#ifdef _UNICODE
#define _tstring std::wstring
#else
#define _tstring std::string
#endif
#endif


LPCTSTR aszExtensions[EXTENSIONS_COUNT] = { TEXT(".jpg"), TEXT(".jpeg"), TEXT(".png"), TEXT(".bmp") };
LPCTSTR aszJacketFileNames[JACKET_FILENAME_COUNT] =
{
	TEXT("Folder"),	TEXT("Jacket"),
	TEXT("Covers"), TEXT("Artwork"), TEXT("Sans"), TEXT("Pictures"), TEXT("Images"), TEXT("Photos"),
	TEXT("Cover"), TEXT("Art"), TEXT("Scan"), TEXT("Picture"), TEXT("Image"), TEXT("Photo"),
};

struct AlbumJacketData
{
	Gdiplus::Bitmap *bmp;
	_tstring stLastFolder;
	HWND hwndProperty;
	BYTE bCurrentAlpha;
	BYTE bAlphaNormal;
	BYTE bAlphaPushed;
	BYTE bAlphaOver;
};


extern "C"
{
	void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info);
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks);
void WINAPI UninitializePlugin(LPVOID lpvUserData);
HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner);
int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
int WINAPI OnSoundOpen(LPVOID lpvUserData, LPVOID lpvInstance, const GenericPlugin::SoundInfo& Info);
int WINAPI OnSoundChange(LPVOID lpvUserData, LPVOID lpvInstance, const GenericPlugin::SoundInfo& Info);
int WINAPI OnSoundClose(LPVOID lpvUserData, LPVOID lpvInstance);
BOOL WINAPI FaceItemNotifyCallback(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFieldName, GenericPlugin::FaceItemNotify NotifyCode, LPCVOID pNotifyStruct);

static void DrawJacketAll(AlbumJacketData *data);
static void DrawJacket(HANDLE hFace, AlbumJacketData *data);
static void DrawJacketSub(Gdiplus::Bitmap *bmp, const GenericPlugin::DrawItem& Item, BYTE bConstAlpha);
static void LoadJacket(AlbumJacketData *data, LPCTSTR pszFolder, LPCTSTR pszSrcFileName);

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam);

GenericPlugin::GenericPluginInfo g_Info =
{
	0x200,
	TEXT("SampleAlbumJacket"),
	TEXT("Northern Verse (Project9k)"),
	TEXT("tFCX`vOC̃Tvł"),
	InitializePlugin,
	UninitializePlugin,
	ShowProperty,
	OnFaceLoad,
	NULL,//OnFaceUnload,
	OnSoundOpen,
	OnSoundChange,
	OnSoundClose,
	NULL,//OnUpdateVisual,
	NULL,//OnSoundStatus
	NULL,//FaceEffectCallback
	FaceItemNotifyCallback,
};

GenericPlugin::PluginCallbacks *g_callbacks = NULL;
HINSTANCE g_hThisInstance = NULL;

BOOL APIENTRY DllMain(HINSTANCE hThisInstance, DWORD dwNotification, LPVOID lpReserved)
{
	if (dwNotification == DLL_PROCESS_ATTACH)
	{
		g_hThisInstance = hThisInstance;
		DisableThreadLibraryCalls(hThisInstance);
	}
	else if (dwNotification == DLL_PROCESS_DETACH)
	{
	}

	return TRUE;
}



void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info)
{
	info = g_Info;
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks)
{
	AlbumJacketData *data = new AlbumJacketData;
	data->bmp = NULL;
	data->stLastFolder.clear();
	data->hwndProperty = NULL;

	g_callbacks = (GenericPlugin::PluginCallbacks *) callbacks;

	return data;
}

void WINAPI UninitializePlugin(LPVOID lpvUserData)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	if (data)
	{
		if (data->bmp)
		{
			delete data->bmp;
		}
		delete data;
	}
}

HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	if (data)
	{
		if (data->hwndProperty)
		{
			SetForegroundWindow(data->hwndProperty);
			return data->hwndProperty;
		}
		else
		{
			data->hwndProperty = CreateDialogParam(g_hThisInstance, MAKEINTRESOURCE(IDD_PROPERTY), hwndOwner, PropertyDialogProc, (LPARAM) data);
			return data->hwndProperty;
		}
	}
	return NULL;
}

int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	data->bAlphaNormal = GetPrivateProfileInt(TARGET_FRAMEBUFFER, TEXT("NormalAlpha"), 255, pszFileName);
	data->bAlphaPushed = GetPrivateProfileInt(TARGET_FRAMEBUFFER, TEXT("PushedAlpha"), 255, pszFileName);
	data->bAlphaOver = GetPrivateProfileInt(TARGET_FRAMEBUFFER, TEXT("MouseOverAlpha"), 255, pszFileName);
	data->bCurrentAlpha = data->bAlphaNormal;
	DrawJacket(hFace, data);

	return 0;
}

int WINAPI OnSoundOpen(LPVOID lpvUserData, LPVOID lpvInstance, const GenericPlugin::SoundInfo& Info)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	LoadJacket(data, Info.pszFilePath, Info.pszFileName);
	DrawJacketAll(data);

	return 0;
}

int WINAPI OnSoundChange(LPVOID lpvUserData, LPVOID lpvInstance, const GenericPlugin::SoundInfo& Info)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	LoadJacket(data, Info.pszFilePath, Info.pszFileName);
	DrawJacketAll(data);

	return 0;
}

int WINAPI OnSoundClose(LPVOID lpvUserData, LPVOID lpvInstance)
{
	AlbumJacketData *data = (AlbumJacketData *) lpvUserData;

	if (data->bmp)
	{
		delete data->bmp;
		data->bmp = NULL;
	}

	DrawJacketAll(data);

	return 0;
}

static void DrawJacketAll(AlbumJacketData *data)
{
	if (g_callbacks && data)
	{
		int i;
		for (i = 0; ; i++)
		{
			HANDLE hFace = g_callbacks->GetFaceFromIndex(i);
			if (hFace)
			{
				DrawJacket(hFace, data);
			}
			else
			{
				break;
			}
		}
	}
}

static void DrawJacket(HANDLE hFace, AlbumJacketData *data)
{
	if (g_callbacks && data)
	{
		GenericPlugin::DrawItem Item;
		g_callbacks->WaitLockFace(hFace, 0, NULL, INFINITE);
		Item = g_callbacks->GetFaceItem(hFace, TARGET_FRAMEBUFFER);
		DrawJacketSub(data->bmp, Item, data->bCurrentAlpha);
		g_callbacks->InvalidateFaceRect(hFace, Item.rcView);
		g_callbacks->UpdateFace(hFace);
		g_callbacks->UnlockFace(hFace);
	}
}

static void DrawJacketSub(Gdiplus::Bitmap *bmp, const GenericPlugin::DrawItem& Item, BYTE bConstAlpha)
{
	if (Item.FrameBuffer)
	{
		Gdiplus::Graphics gr(Item.FrameBuffer);

		gr.Clear(0);

		if (bmp)
		{
			int DestWidth = Item.FrameBuffer->GetWidth();
			int DestHeight = Item.FrameBuffer->GetHeight();
			int SrcWidth = bmp->GetWidth();
			int SrcHeight = bmp->GetHeight();

			Gdiplus::Rect rcDest(0, 0, DestWidth, DestHeight);
			if (SrcWidth <= DestWidth && SrcHeight <= DestHeight)
			{
				rcDest.Width = SrcWidth;
				rcDest.X = (DestWidth - SrcWidth) / 2;
				rcDest.Height = SrcHeight;
				rcDest.Y = (DestHeight - SrcWidth) / 2;
			}
			else
			{
				int y = SrcHeight * DestWidth / SrcWidth;
				if (y > DestHeight)
				{
					rcDest.Width = DestHeight * SrcWidth / SrcHeight;
					rcDest.X = (DestWidth - rcDest.Width) / 2;
				}
				else
				{
					rcDest.Height = y;
					rcDest.Y = (DestHeight - rcDest.Height) / 2;
				}
			}
			gr.SetCompositingMode(Gdiplus::CompositingModeSourceCopy);
			gr.SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
			gr.SetInterpolationMode(Gdiplus::InterpolationModeHighQuality);
			Gdiplus::ImageAttributes attr;
			const Gdiplus::ColorMatrix cmx = {
				1.0f, 0.0f, 0.0f, 0.0f, 0.0f,   // Red
				0.0f, 1.0f, 0.0f, 0.0f, 0.0f,   // Green
				0.0f, 0.0f, 1.0f, 0.0f, 0.0f,   // Blue
				0.0f, 0.0f, 0.0f, bConstAlpha / 255.0f, 0.0f,   // Alpha
				0.0f, 0.0f, 0.0f, 0.0f, 1.0f    // must be 1
			};
			attr.SetColorMatrix(&cmx, Gdiplus::ColorMatrixFlagsDefault, Gdiplus::ColorAdjustTypeBitmap); // Set ColorMatrix
			gr.DrawImage(bmp, rcDest, 0, 0, SrcWidth, SrcHeight, Gdiplus::UnitPixel, &attr, NULL, NULL);
		}
	}
}

static void LoadJacket(AlbumJacketData *data, LPCTSTR pszFolder, LPCTSTR pszSrcFileName)
{
	int i;

	if (data->bmp)
	{
		delete data->bmp;
		data->bmp = NULL;
	}

	for (i = 0; i < EXTENSIONS_COUNT; i++)
	{
		HANDLE hffd;
		WIN32_FIND_DATA ffd;
		TCHAR szBuf[STR_LEN];

		_tcscpy_s(szBuf, STR_LEN, pszFolder);
		PathAppend(szBuf, pszSrcFileName);
		PathRenameExtension(szBuf, aszExtensions[i]);

		hffd = FindFirstFile(szBuf, &ffd);
		if (hffd && hffd != INVALID_HANDLE_VALUE)
		{
			FindClose(hffd);
			_tcscpy_s(szBuf, STR_LEN, pszFolder);
			PathAppend(szBuf, ffd.cFileName);

#ifdef _UNICODE
			data->bmp = new Gdiplus::Bitmap(szBuf);
#else
			int len = MultiByteToWideChar(CP_ACP, 0, szBuf, -1, NULL, 0);
			LPWSTR pwsz = (LPWSTR) _malloca(len * sizeof(WCHAR));
			MultiByteToWideChar(CP_ACP, 0, szBuf, -1, pwsz, len);
			data->bmp = new Gdiplus::Bitmap(pwsz);
			_freea(pwsz);
#endif
			if (data->bmp)
			{
				if (data->bmp->GetLastStatus() != Gdiplus::Ok)
				{
					delete data->bmp;
					data->bmp = NULL;
				}
				else
				{
					return;
				}
			}
		}
	}

	int j;
	for (j = 0; j < JACKET_FILENAME_COUNT; j++)
	{
		for (i = 0; i < EXTENSIONS_COUNT; i++)
		{
			HANDLE hffd;
			WIN32_FIND_DATA ffd;
			TCHAR szBuf[STR_LEN];

			_tcscpy_s(szBuf, STR_LEN, pszFolder);
			PathAppend(szBuf, aszJacketFileNames[j]);
			PathAddExtension(szBuf, aszExtensions[i]);

			hffd = FindFirstFile(szBuf, &ffd);
			if (hffd && hffd != INVALID_HANDLE_VALUE)
			{
				FindClose(hffd);
				_tcscpy_s(szBuf, STR_LEN, pszFolder);
				PathAppend(szBuf, ffd.cFileName);

	#ifdef _UNICODE
				data->bmp = new Gdiplus::Bitmap(szBuf);
	#else
				int len = MultiByteToWideChar(CP_ACP, 0, szBuf, -1, NULL, 0);
				LPWSTR pwsz = (LPWSTR) _malloca(len * sizeof(WCHAR));
				MultiByteToWideChar(CP_ACP, 0, szBuf, -1, pwsz, len);
				data->bmp = new Gdiplus::Bitmap(pwsz);
				_freea(pwsz);
	#endif
				if (data->bmp)
				{
					if (data->bmp->GetLastStatus() != Gdiplus::Ok)
					{
						delete data->bmp;
						data->bmp = NULL;
					}
					else
					{
						return;
					}
				}
			}
		}
}
	return;
}

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	switch (uMessage)
	{
	case WM_INITDIALOG:
		{
			AlbumJacketData *data = (AlbumJacketData *) lParam;
			SetProp(hwnd, PROP_ALBUMJACKETDATA, (HANDLE) data);
			g_callbacks->RegisterDialog(hwnd);
			ShowWindow(hwnd, SW_SHOWNORMAL);
		}
		break;
	case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
			case IDOK:
			case IDCANCEL:
				{
					DestroyWindow(hwnd);
				}
				break;
			}
		}
		break;
	case WM_DESTROY:
		{
			AlbumJacketData *data = (AlbumJacketData *) GetProp(hwnd, PROP_ALBUMJACKETDATA);
			g_callbacks->UnregisterDialog(hwnd);
			if (data)
			{
				data->hwndProperty = NULL;
			}
		}
		break;
	}

	return FALSE;
}

BOOL WINAPI FaceItemNotifyCallback(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFieldName, GenericPlugin::FaceItemNotify NotifyCode, LPCVOID pNotifyStruct)
{
	if (lpvUserData)
	{
		if (_tcsicmp(pszFieldName, TARGET_FRAMEBUFFER) == 0)
		{
			if (NotifyCode == GenericPlugin::FaceItemNotifySizeChanged)
			{
				const GenericPlugin::FaceItemNotifySizeChangedStruct * p = (const GenericPlugin::FaceItemNotifySizeChangedStruct *) pNotifyStruct;
				DrawJacketSub(((AlbumJacketData *) lpvUserData)->bmp, p->Item, ((AlbumJacketData *) lpvUserData)->bCurrentAlpha);
			}
			else if (NotifyCode == GenericPlugin::FaceItemNotifyMouseEvent)
			{
				const GenericPlugin::FaceItemNotifyMouseEventStruct * p = (const GenericPlugin::FaceItemNotifyMouseEventStruct *) pNotifyStruct;
				BYTE bPrevAlpha = ((AlbumJacketData *) lpvUserData)->bCurrentAlpha;
				switch (p->mouseEvent)
				{
				case GenericPlugin::MouseEventNormal:
					{
						((AlbumJacketData *) lpvUserData)->bCurrentAlpha = ((AlbumJacketData *) lpvUserData)->bAlphaNormal;
					}
					break;
				case GenericPlugin::MouseEventPushed:
					{
						((AlbumJacketData *) lpvUserData)->bCurrentAlpha = ((AlbumJacketData *) lpvUserData)->bAlphaPushed;
					}
					break;
				case GenericPlugin::MouseEventOver:
					{
						((AlbumJacketData *) lpvUserData)->bCurrentAlpha = ((AlbumJacketData *) lpvUserData)->bAlphaOver;
					}
					break;
				}

				if (bPrevAlpha != ((AlbumJacketData *) lpvUserData)->bCurrentAlpha)
				{
					DrawJacketSub(((AlbumJacketData *) lpvUserData)->bmp, p->Item, ((AlbumJacketData *) lpvUserData)->bCurrentAlpha);
				}
			}
			return TRUE;
		}
	}

	return FALSE;
}