#include "..\\uLilith\\Plugin\\GenericPlugin.h"
#include <shlwapi.h>
#include "Resource.h"
#include <process.h>

#include <string>
#include <xstring>
#include <vector>
#include <map>

#include <wincodec.h>
#include <wincodecsdk.h>

#include <math.h>

#include "StretchBitmap.h"


#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "gdiplus.lib")
#pragma comment(lib, "winmm.lib")

#ifdef __INTEL_COMPILER
#pragma warning(disable:1028)
#pragma warning(disable:1738)
#pragma warning(disable:1744)
#endif

#define TARGET_SUBCATEGORY TEXT("Animation")
#define STR_LEN 1024
#define PROP_ANIMATIONDATA TEXT("Prop::AnimationData")

#define LARGE_BUFFER_LEN 65536

#define _USE_DIRECT2D_INTERNAL_

#ifdef _USE_SCALING_LINEAR_
#define _SCALING_INTERPOLATION_MODE_ D2D1_BITMAP_INTERPOLATION_MODE_LINEAR
#else
#define _SCALING_INTERPOLATION_MODE_ D2D1_BITMAP_INTERPOLATION_MODE_NEAREST_NEIGHBOR
#endif

GenericPlugin::PluginCallbacks *g_callbacks = NULL;

namespace std
{
	class string_less_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class wstring_less_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class string_greater_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) > 0);
		}
	};
	class wstring_greater_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) > 0);
		}
	};
};
#ifdef _UNICODE
#define _tstring std::wstring
#define _tstring_less_i std::wstring_less_i
#else
#define _tstring std::string
#define _tstring_less_i std::string_less_i
#endif



inline BOOL NVStringToBool(LPCTSTR str)
{
	if (_tcsicmp(str, TEXT("TRUE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("ENABLE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("On")) == 0)
	{
		return TRUE;
	}
	else if (_tstoi(str) == 1)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
inline BOOL GetPrivateProfileBool(LPCTSTR pszField, LPCTSTR pszKey, BOOL fDefault, LPCTSTR pszFileName)
{
	TCHAR szBuf[256];

	GetPrivateProfileString(pszField, pszKey, fDefault ? TEXT("TRUE") : TEXT("FALSE"), szBuf, 256, pszFileName);
	return NVStringToBool(szBuf);
}


struct AnimationFrame
{
	int nIndexMin;
	int nIndexMax;
	int nFrameTimeMin;
	int nFrameTimeMax;
};
typedef std::vector<AnimationFrame> AnimationFrameList;


class AnimationEntry
{
public:
	AnimationEntry(HANDLE hFace);
	~AnimationEntry();
#ifdef _USE_DIRECT2D_INTERNAL_
	BOOL Load(HANDLE hFace, LPCTSTR pszFieldName, LPCTSTR pszFaceFileName, IWICImagingFactory *pImagingFactory);
#else
	BOOL Load(LPCTSTR pszFieldName, LPCTSTR pszFaceFileName);
#endif
	void Unload(void);
	void DrawFrameSub(const GenericPlugin::DrawItem& Item);
	void DrawFrameSubD2D(HANDLE hFace, const GenericPlugin::DrawItemD2D& Item);
	bool SetMouseEvent(GenericPlugin::MouseEvent mouseEvent);

protected:
	static unsigned int __stdcall ThreadProc(void *lpvParam);
	unsigned int ThreadProcSub(void);
	unsigned int ThreadProcSubCustomSequence(void);
	BOOL DrawFrame(BOOL fForce = FALSE);
#ifdef _USE_DIRECT2D_INTERNAL_
	void ExtractFrame(HANDLE hFace, UINT uIndex);
	BOOL CreateRenderTarget(HANDLE hFace, UINT32 uWidth, UINT32 uHeight);
#endif

	_tstring stFieldName;
	Gdiplus::Bitmap *bmp;
	SIZE sizFrame;
	int nXCount;
	int nYCount;
	int nTotalCount;
	int nCurrentIndex;
	int nInterval;
	BOOL fIsAnimationImage;
	HANDLE hThread;
	HANDLE hKillEvent;
	GUID guidFrameDimension;
	int nLoopCount;
	int nCurrentLoopCount;
	BOOL fIsUseCustomSequence;
	AnimationFrameList frameList;
	HANDLE hFace;
#ifdef _USE_DIRECT2D_INTERNAL_
	IWICBitmapDecoder *pDecoder;
	ID2D1BitmapRenderTarget *pTarget;
	IWICImagingFactory *pImagingFactory;
	int nCurrentFrameInBuffer;
#endif
	BOOL fIsFrameBufferInitialized;
	BOOL fIsResetOnRedraw;
	BYTE bCurrentAlpha;
	BYTE bAlphaNormal;
	BYTE bAlphaPushed;
	BYTE bAlphaOver;
	bool fNeedColorConvert;
	DWORD dwViewScale;
};
typedef std::map<_tstring, AnimationEntry *, _tstring_less_i> AnimationList;
typedef std::map<HANDLE, AnimationList> AnimationFaceList;
struct AnimationData
{
	AnimationFaceList FaceList;
	HWND hwndProperty;
#ifdef _USE_DIRECT2D_INTERNAL_
	IWICImagingFactory *pImagingFactory;
#endif
	void ClearAnimationListAll(void)
	{
		AnimationFaceList::iterator itList;

		for (itList = FaceList.begin(); itList != FaceList.end(); )
		{
			AnimationList::iterator it;
			for (it = itList->second.begin(); it != itList->second.end(); )
			{
				if (it->second)
				{
					delete (it->second);
				}
				it = itList->second.erase(it);
			}
			itList = FaceList.erase(itList);
		}
	}
	void ClearAnimationList(HANDLE hFace)
	{
		AnimationFaceList::iterator itList;

		itList = FaceList.find(hFace);
		if (itList != FaceList.end())
		{
			AnimationList::iterator it;
			for (it = itList->second.begin(); it != itList->second.end(); )
			{
				if (it->second)
				{
					delete (it->second);
				}
				it = itList->second.erase(it);
			}
		}
	}
};

extern "C"
{
	void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info);
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks);
void WINAPI UninitializePlugin(LPVOID lpvUserData);
HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner);
int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
BOOL WINAPI FaceItemNotifyCallback(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFieldName, GenericPlugin::FaceItemNotify NotifyCode, LPCVOID pNotifyStruct);

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam);

GenericPlugin::GenericPluginInfo g_Info =
{
	0x200,
	TEXT("SampleAnimationFace"),
	TEXT("Northern Verse (Project9k)"),
	TEXT("tFCX`vOC̃Tvł"),
	InitializePlugin,
	UninitializePlugin,
	ShowProperty,
	OnFaceLoad,
	OnFaceUnload,
	NULL,//OnSoundOpen,
	NULL,//OnSoundChange,
	NULL,//OnSoundClose,
	NULL,//OnUpdateVisual,
	NULL,//OnSoundStatus
	NULL,//FaceEffectCallback
	FaceItemNotifyCallback,
	GenericPlugin::SupportedDrawModeDirect2D,
};

HINSTANCE g_hThisInstance = NULL;

BOOL APIENTRY DllMain(HINSTANCE hThisInstance, DWORD dwNotification, LPVOID lpReserved)
{
	if (dwNotification == DLL_PROCESS_ATTACH)
	{
		g_hThisInstance = hThisInstance;
		DisableThreadLibraryCalls(hThisInstance);
	}
	else if (dwNotification == DLL_PROCESS_DETACH)
	{
	}

	return TRUE;
}



void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info)
{
	info = g_Info;
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks)
{
#ifdef _USE_DIRECT2D_INTERNAL_
	IWICImagingFactory *pImagingFactory = NULL;
	HRESULT hres = CoCreateInstance(CLSID_WICImagingFactory, NULL, CLSCTX_INPROC_SERVER, __uuidof(IWICImagingFactory), (LPVOID *)&pImagingFactory);
	if (FAILED(hres))
	{
		if (callbacks->GetDrawMode() == GenericPlugin::DrawModeDirect2D)
		{
			return NULL;
		}
	}
#endif
	AnimationData *data = new AnimationData;
	data->FaceList.clear();
	data->hwndProperty = NULL;
#ifdef _USE_DIRECT2D_INTERNAL_
	data->pImagingFactory = pImagingFactory;
#endif

	g_callbacks = (GenericPlugin::PluginCallbacks *) callbacks;

	return data;
}

void WINAPI UninitializePlugin(LPVOID lpvUserData)
{
	AnimationData *data = (AnimationData *) lpvUserData;

	if (data)
	{
		data->ClearAnimationListAll();
#ifdef _USE_DIRECT2D_INTERNAL_
		if (data->pImagingFactory)
		{
			data->pImagingFactory->Release();
		}
#endif
		delete data;
	}
}

HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner)
{
	AnimationData *data = (AnimationData *) lpvUserData;

	if (data)
	{
		if (data->hwndProperty)
		{
			SetForegroundWindow(data->hwndProperty);
			return data->hwndProperty;
		}
		else
		{
			data->hwndProperty = CreateDialogParam(g_hThisInstance, MAKEINTRESOURCE(IDD_PROPERTY), hwndOwner, PropertyDialogProc, (LPARAM) data);
			return data->hwndProperty;
		}
	}
	return NULL;
}

int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	AnimationData *data = (AnimationData *) lpvUserData;

	if (data)
	{
		AnimationFaceList::iterator it;

		it = data->FaceList.find(hFace);
		if (it == data->FaceList.end())
		{
			data->FaceList.insert(std::make_pair(hFace, AnimationList()));
			it = data->FaceList.find(hFace);
		}

		if (it != data->FaceList.end())
		{
			TCHAR szFieldName[STR_LEN];
			HANDLE hffi = g_callbacks->FindFirstFaceItem(hFace, TEXT("Animation"), szFieldName, STR_LEN);
			if (hffi)
			{
				do
				{
					AnimationEntry *entry = new AnimationEntry(hFace);

#ifdef _USE_DIRECT2D_INTERNAL_
					if (entry->Load(hFace, szFieldName, pszFileName, data->pImagingFactory))
#else
					if (entry->Load(szFieldName, pszFileName))
#endif
					{
						it->second.insert(std::make_pair(szFieldName, entry));
					}
					else
					{
						delete entry;
					}
				}
				while (g_callbacks->FindNextFaceItem(hffi, szFieldName, STR_LEN));
			}
			g_callbacks->FindFaceItemClose(hffi);
		}
	}

	return 0;
}

int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	AnimationData *data = (AnimationData *) lpvUserData;

	if (data)
	{
		data->ClearAnimationList(hFace);
	}

	return 0;
}


AnimationEntry::AnimationEntry(HANDLE hFace)
{
	stFieldName.clear();
	bmp = NULL;
	sizFrame.cx = 0;
	sizFrame.cy = 0;
	nXCount = 0;
	nYCount = 0;
	nTotalCount = 0;
	nCurrentIndex = 0;
	nInterval = 0;
	hThread = NULL;
	hKillEvent = NULL;
	fIsAnimationImage = FALSE;
	nLoopCount = 0;
	nCurrentLoopCount = 0;
	frameList.clear();
	this->hFace = hFace;
#ifdef _USE_DIRECT2D_INTERNAL_
	pDecoder = NULL;
	pTarget = NULL;
	pImagingFactory = NULL;
	nCurrentFrameInBuffer = -1;
#endif
	fIsFrameBufferInitialized = FALSE;
	fIsResetOnRedraw = FALSE;
	bCurrentAlpha = 255;
	bAlphaNormal = 255;
	bAlphaPushed = 255;
	bAlphaOver = 255;
	fNeedColorConvert = false;
	dwViewScale = 100;
}

AnimationEntry::~AnimationEntry()
{
	Unload();
}
int GetEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
   UINT  num = 0;          // number of image encoders
   UINT  size = 0;         // size of the image encoder array in bytes

   Gdiplus::ImageCodecInfo* pImageCodecInfo = NULL;

   Gdiplus::GetImageEncodersSize(&num, &size);
   if(size == 0)
      return -1;  // Failure

   pImageCodecInfo = (Gdiplus::ImageCodecInfo*)(malloc(size));
   if(pImageCodecInfo == NULL)
      return -1;  // Failure

   GetImageEncoders(num, size, pImageCodecInfo);

   for(UINT j = 0; j < num; ++j)
   {
      if( wcscmp(pImageCodecInfo[j].MimeType, format) == 0 )
      {
         *pClsid = pImageCodecInfo[j].Clsid;
         free(pImageCodecInfo);
         return j;  // Success
      }    
   }

   free(pImageCodecInfo);
   return -1;  // Failure
}



#ifdef _USE_DIRECT2D_INTERNAL_
BOOL AnimationEntry::Load(HANDLE hFace, LPCTSTR pszFieldName, LPCTSTR pszFaceFileName, IWICImagingFactory *pImagingFactory)
#else
BOOL AnimationEntry::Load(LPCTSTR pszFieldName, LPCTSTR pszFaceFileName)
#endif
{
#ifdef _USE_DIRECT2D_INTERNAL_
	if (pImagingFactory)
	{
		pImagingFactory->AddRef();
		this->pImagingFactory = pImagingFactory;
	}
#endif
	TCHAR szFaceDirectory[STR_LEN];
	TCHAR szBuf[STR_LEN];
	WCHAR szFileName[STR_LEN];

	_tcsncpy_s(szFaceDirectory, STR_LEN, pszFaceFileName, STR_LEN - 1);
	PathRemoveFileSpec(szFaceDirectory);
	GetPrivateProfileString(pszFieldName, TEXT("AnimationImageFile"), TEXT(""), szBuf, STR_LEN, pszFaceFileName);
	if (! szBuf[0])
	{
		return FALSE;
	}

	if (PathIsRelative(szBuf))
	{
#ifdef _UNICODE
		_tcscpy_s(szFileName, STR_LEN, pszFaceFileName);
		PathRemoveFileSpec(szFileName);
		PathAppend(szFileName, szBuf);
#else
		TCHAR szTemp[STR_LEN];
		_tcscpy_s(szTemp, STR_LEN, pszFaceFileName);
		PathRemoveFileSpec(szTemp);
		PathAppend(szTemp, szBuf);
		MultiByteToWideChar(CP_ACP, 0, szTemp, -1, szFileName, STR_LEN);
#endif
	}
	else
	{
#ifdef _UNICODE
		_tcscpy_s(szFileName, STR_LEN, szBuf);
#else
		MultiByteToWideChar(CP_ACP, 0, szBuf, -1, szFileName, STR_LEN);
#endif
	}

	fIsUseCustomSequence = GetPrivateProfileBool(pszFieldName, TEXT("UseCustomSequence"), FALSE, pszFaceFileName);
	fIsResetOnRedraw = GetPrivateProfileBool(pszFieldName, TEXT("ResetOnRedraw"), FALSE, pszFaceFileName);

	bAlphaNormal = GetPrivateProfileInt(pszFieldName, TEXT("NormalAlpha"), 255, pszFaceFileName);
	bAlphaPushed = GetPrivateProfileInt(pszFieldName, TEXT("PushedAlpha"), 255, pszFaceFileName);
	bAlphaOver = GetPrivateProfileInt(pszFieldName, TEXT("MouseOverAlpha"), 255, pszFaceFileName);
	bCurrentAlpha = bAlphaNormal;

	dwViewScale = 100;
#ifdef _USE_DIRECT2D_INTERNAL_
	if (g_callbacks->GetDrawMode() == GenericPlugin::DrawModeDirect2D)
	{
		if (! pImagingFactory)
		{
			Unload();
			return FALSE;
		}
		HRESULT hres;
		hres = pImagingFactory->CreateDecoderFromFilename(szFileName, NULL, GENERIC_READ, WICDecodeMetadataCacheOnLoad, &pDecoder);
		if (pDecoder)
		{
			UINT uWidth = 0, uHeight = 0;
			IWICBitmapFrameDecode * pFrame = NULL;
			hres = pDecoder->GetFrame(0, &pFrame);
			if (pFrame)
			{
				UINT uCount = 0;
				pDecoder->GetFrameCount(&uCount);
				nTotalCount = uCount;

				pFrame->GetSize(&uWidth, &uHeight);
				if (uWidth <= 0 || uHeight <= 0)
				{
					pFrame->Release();
					Unload();
					return FALSE;
				}

				nInterval = 0;
				{
					IWICMetadataQueryReader *pIMetadataQueryReader = NULL;
					pFrame->GetMetadataQueryReader(&pIMetadataQueryReader);
					if (pIMetadataQueryReader)
					{
						PROPVARIANT prop;
						PropVariantInit(&prop);
						HRESULT hres = pIMetadataQueryReader->GetMetadataByName(L"/grctlext/Delay", &prop);
						if (SUCCEEDED(hres) && prop.vt == VT_UI2)
						{
							nInterval = prop.uiVal * 10;
						}
						pIMetadataQueryReader->Release();

						PropVariantClear(&prop);
					}
				}
				pFrame->Release();
			}
			if (nTotalCount > 1)
			{
				{
					IWICMetadataQueryReader *pIMetadataQueryReader = NULL;
					HRESULT hres = pDecoder->GetMetadataQueryReader(&pIMetadataQueryReader);
					if (pIMetadataQueryReader)
					{
						PROPVARIANT prop;
						PropVariantInit(&prop);
						HRESULT hres = pIMetadataQueryReader->GetMetadataByName(L"/appext/application", &prop);
						if (SUCCEEDED(hres) && prop.vt == (VT_UI1 | VT_VECTOR) && prop.caub.cElems == 11 &&
							(memcmp(prop.caub.pElems, "NETSCAPE2.0", prop.caub.cElems) == 0 ||
							memcmp(prop.caub.pElems, "ANIMEXTS1.0", prop.caub.cElems) == 0))
						{
							PropVariantClear(&prop);
							hres = pIMetadataQueryReader->GetMetadataByName(L"/appext/data", &prop);
							if (SUCCEEDED(hres) && prop.vt == (VT_UI1 | VT_VECTOR) && prop.caub.cElems >= 4 &&
								prop.caub.pElems[0] > 0 && prop.caub.pElems[1] == 1)
							{
								nLoopCount = MAKEWORD(prop.caub.pElems[2], prop.caub.pElems[3]);
							}
						}
						pIMetadataQueryReader->Release();
					}
				}

				fIsAnimationImage = TRUE;

				sizFrame.cx = uWidth;
				nXCount = 1;
				sizFrame.cy = uHeight;
				nYCount = 1;

				if (! fIsUseCustomSequence)
				{
					if (nInterval <= 0)
					{
						nInterval = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), 0, pszFaceFileName);
						if (nInterval <= 0)
						{
							Unload();
							return FALSE;
						}
					}
					else
					{
						int t = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), -1, pszFaceFileName);
						if (t > 0)
						{
							nInterval = t;
						}
					}
				}
				nLoopCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationLoopCount"), nLoopCount, pszFaceFileName);

				if (! CreateRenderTarget(hFace, sizFrame.cx, sizFrame.cy))
				{
					Unload();
					return FALSE;
				}
				dwViewScale = g_callbacks->GetFaceScale();
				ExtractFrame(hFace, 0);
			}
			else
			{
				DWORD dwScale = g_callbacks->GetFaceScale();
				SIZE sizFrameOrg;
				uWidth = uWidth * dwScale / 100;
				uHeight = uHeight * dwScale / 100;
				nTotalCount = 1;
				sizFrame.cx = GetPrivateProfileInt(pszFieldName, TEXT("AnimationFrameWidth"), 0, pszFaceFileName);
				sizFrameOrg.cx = sizFrame.cx;
				sizFrame.cx = sizFrame.cx * dwScale / 100;
				if (sizFrame.cx > (INT) uWidth || sizFrame.cx <= 0)
				{
					Unload();
					return FALSE;
				}
				nXCount = uWidth / sizFrame.cx;
				sizFrame.cy = GetPrivateProfileInt(pszFieldName, TEXT("AnimationFrameHeight"), 0, pszFaceFileName);
				sizFrameOrg.cy = sizFrame.cy;
				sizFrame.cy = sizFrame.cy * dwScale / 100;
				if (sizFrame.cy > (INT) uHeight || sizFrame.cy <= 0)
				{
					Unload();
					return FALSE;
				}
				nYCount = uHeight / sizFrame.cy;
				nTotalCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationTotalCount"), 0, pszFaceFileName);
				if (nTotalCount <= 0 || nTotalCount > nXCount * nYCount)
				{
					Unload();
					return FALSE;
				}
				if (! fIsUseCustomSequence)
				{
					nInterval = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), 0, pszFaceFileName);
					if (nInterval <= 0)
					{
						Unload();
						return FALSE;
					}
				}
				nLoopCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationLoopCount"), 0, pszFaceFileName);

				if (dwScale != 100)
				{
					IWICBitmapDecoder *pScaledDecoder = NULL;
					IStream *pStream = GetScaledBitmapStream_DLL(szFileName, g_callbacks->GetFaceScale(), &sizFrameOrg);
					if (pStream)
					{
						hres = pImagingFactory->CreateDecoderFromStream(pStream, NULL, WICDecodeMetadataCacheOnLoad, &pScaledDecoder);
						if (pScaledDecoder)
						{
							pDecoder->Release();
							pDecoder = pScaledDecoder;
							IWICBitmapFrameDecode * pFrame = NULL;
							hres = pDecoder->GetFrame(0, &pFrame);
							if (pFrame)
							{
								pFrame->GetSize(&uWidth, &uHeight);
								pFrame->Release();
							}
						}
						pStream->Release();
					}
				}
				if (! CreateRenderTarget(hFace, uWidth, uHeight))
				{
					BOOL fSuccess = FALSE;
					if (((double) uWidth / uHeight) < 2.0 || ((double) uHeight / uWidth) < 2.0)
					{
						int sq = (int) ceil(sqrt((double) uWidth * uHeight));
						int nNewX = (sq + sizFrame.cx - 1) / sizFrame.cx;
						int nNewY = (nTotalCount + nNewX - 1) / nNewX;

						while (nNewX * nNewY < nTotalCount)
						{
							nNewY++;
						}
						Gdiplus::Bitmap *bmp = GetScaledBitmap_DLL(szFileName, g_callbacks->GetFaceScale(), &sizFrameOrg);
						if (bmp->GetLastStatus() == Gdiplus::Ok)
						{
							Gdiplus::Bitmap tmp(sizFrame.cx * nNewX, sizFrame.cy * nNewY, PixelFormat32bppPARGB);
							{
								int i, j, k;
								Gdiplus::Graphics gr(&tmp);
								gr.SetPageUnit(Gdiplus::UnitPixel);
								gr.SetCompositingMode(Gdiplus::CompositingModeSourceCopy);
								gr.SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
								gr.SetInterpolationMode(Gdiplus::InterpolationModeNearestNeighbor);

								for (i = 0, k = 0; i < nNewY && k < nTotalCount; i++)
								{
									for (j = 0; j < nNewX && k < nTotalCount; j++, k++)
									{
										gr.DrawImage(bmp, 
											Gdiplus::Rect(j* sizFrame.cx, i * sizFrame.cy, sizFrame.cx, sizFrame.cy),
											(k % nXCount) * sizFrame.cx, (k / nXCount) * sizFrame.cy, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
									}
								}
							}
							if (CreateRenderTarget(hFace, sizFrame.cx * nNewX, sizFrame.cy * nNewY))
							{
								Gdiplus::BitmapData data;
								Gdiplus::Rect rct(0,0, tmp.GetWidth(), tmp.GetHeight());
								if (Gdiplus::Ok == tmp.LockBits(&rct, Gdiplus::ImageLockModeRead, PixelFormat32bppPARGB, &data))
								{
									D2D1_PIXEL_FORMAT format = D2D1::PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, D2D1_ALPHA_MODE_PREMULTIPLIED);
									ID2D1Bitmap *pD2Bmp = NULL;
									if (SUCCEEDED(pTarget->GetBitmap(&pD2Bmp)))
									{
										pD2Bmp->CopyFromMemory(NULL, data.Scan0, data.Stride);
										pD2Bmp->Release();
										fSuccess = TRUE;
									}
									tmp.UnlockBits(&data);
								}
								nXCount = nNewX;
								nYCount = nNewY;
							}
						}
						delete bmp;
					}
					if (! fSuccess)
					{
						Unload();
						return FALSE;
					}
				}
				else
				{
					ExtractFrame(hFace, 0);
				}
			}
		}
	}
	else
#endif
	{
		bmp = new Gdiplus::Bitmap(szFileName);
		if (! bmp)
		{
			return FALSE;
		}
		else if (bmp->GetLastStatus() != Gdiplus::Ok)
		{
			Unload();
			return FALSE;
		}
		else if (bmp->GetWidth() <= 0 || bmp->GetHeight() <= 0)
		{
			Unload();
			return FALSE;
		}

		{
			UINT count = 0;
			count = bmp->GetFrameDimensionsCount();
			GUID *pDimensionIDs = new GUID[count];
			bmp->GetFrameDimensionsList(pDimensionIDs, count);
			guidFrameDimension = pDimensionIDs[0];
			nTotalCount = bmp->GetFrameCount(&guidFrameDimension);
			delete [] pDimensionIDs;

		}
		if (nTotalCount > 1)
		{
			{	//  Windows XP Animation GIF + alpha ΍
				OSVERSIONINFO os;
				GetVersionEx(&os);
				if (os.dwMajorVersion <= 5)
				{
					fNeedColorConvert = true;
				}
			}

			fIsAnimationImage = TRUE;

			sizFrame.cx = bmp->GetWidth();
			nXCount = 1;
			sizFrame.cy = bmp->GetHeight();
			nYCount = 1;

			{
				int nSize = bmp->GetPropertyItemSize(PropertyTagFrameDelay);
				Gdiplus::PropertyItem * prop = (Gdiplus::PropertyItem *) _malloca(nSize);
				if (Gdiplus::Ok == bmp->GetPropertyItem(PropertyTagFrameDelay, nSize, prop))
				{
					switch (prop->type)
					{
					case PropertyTagTypeByte:
						{
							nInterval = *((BYTE *) prop->value) * 10;
						}
						break;
					case PropertyTagTypeShort:
						{
							nInterval = *((SHORT *) prop->value) * 10;
						}
						break;
					case PropertyTagTypeLong:
						{
							nInterval = *((LONG *) prop->value) * 10;
						}
						break;
					default:
						{
							_freea(prop);
							Unload();
							return FALSE;
						}
					}
				}
				else
				{
					nInterval = 0;
				}
				_freea(prop);
			}

			{
				int nSize = bmp->GetPropertyItemSize(PropertyTagLoopCount);
				Gdiplus::PropertyItem * prop = (Gdiplus::PropertyItem *) _malloca(nSize);
				if (Gdiplus::Ok == bmp->GetPropertyItem(PropertyTagLoopCount, nSize, prop))
				{
					switch (prop->type)
					{
					case PropertyTagTypeByte:
						{
							nLoopCount = *((BYTE *) prop->value);
						}
						break;
					case PropertyTagTypeShort:
						{
							nLoopCount = *((SHORT *) prop->value);
						}
						break;
					case PropertyTagTypeLong:
						{
							nLoopCount = *((LONG *) prop->value);
						}
						break;
					default:
						{
							nLoopCount = 0;
						}
					}
				}
				else
				{
					nLoopCount = 0;
				}
				_freea(prop);
			}

			if (! fIsUseCustomSequence)
			{
				if (nInterval <= 0)
				{
					nInterval = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), 0, pszFaceFileName);
					if (nInterval <= 0)
					{
						Unload();
						return FALSE;
					}
				}
				else
				{
					int t = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), -1, pszFaceFileName);
					if (t > 0)
					{
						nInterval = t;
					}
				}
			}
			nLoopCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationLoopCount"), nLoopCount, pszFaceFileName);
			dwViewScale = g_callbacks->GetFaceScale();
		}
		else
		{
			DWORD dwScale = g_callbacks->GetFaceScale();
			SIZE sizFrameOrg;

			sizFrame.cx = GetPrivateProfileInt(pszFieldName, TEXT("AnimationFrameWidth"), 0, pszFaceFileName);
			sizFrameOrg.cx = sizFrame.cx;
			sizFrame.cx = sizFrame.cx * g_callbacks->GetFaceScale() / 100;
			if (sizFrameOrg.cx > (int) bmp->GetWidth())
			{
				Unload();
				return FALSE;
			}
			nXCount = bmp->GetWidth() / sizFrameOrg.cx;
			sizFrame.cy = GetPrivateProfileInt(pszFieldName, TEXT("AnimationFrameHeight"), 0, pszFaceFileName);
			sizFrameOrg.cy = sizFrame.cy;
			sizFrame.cy = sizFrame.cy * g_callbacks->GetFaceScale() / 100;
			if (sizFrameOrg.cy > (int) bmp->GetHeight())
			{
				Unload();
				return FALSE;
			}
			nYCount = bmp->GetHeight() / sizFrameOrg.cy;
			nTotalCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationTotalCount"), 0, pszFaceFileName);
			if (nTotalCount <= 0 || nTotalCount > nXCount * nYCount)
			{
				Unload();
				return FALSE;
			}
			if (! fIsUseCustomSequence)
			{
				nInterval = GetPrivateProfileInt(pszFieldName, TEXT("AnimationInterval"), 0, pszFaceFileName);
				if (nInterval <= 0)
				{
					Unload();
					return FALSE;
				}
			}
			nLoopCount = GetPrivateProfileInt(pszFieldName, TEXT("AnimationLoopCount"), 0, pszFaceFileName);

			Gdiplus::Bitmap *scaledBmp = GetScaledBitmap_DLL(szFileName, g_callbacks->GetFaceScale(), &sizFrameOrg);
			if (scaledBmp)
			{
				delete bmp;
				bmp = scaledBmp;
			}

		}
	}

	if (fIsUseCustomSequence)
	{
		LPTSTR pReadBuf = (LPTSTR) _malloca(sizeof(TCHAR) * LARGE_BUFFER_LEN);
		GetPrivateProfileString(pszFieldName, TEXT("CustomSequence"), TEXT(""), pReadBuf, LARGE_BUFFER_LEN, pszFaceFileName);

		if (pReadBuf[0])
		{
			LPTSTR pCur;
			LPTSTR pContext = NULL;

			for (pCur = _tcstok_s(pReadBuf, TEXT(","), &pContext); pCur; pCur = _tcstok_s(NULL, TEXT(","), &pContext))
			{
				BOOL fParseSuccess = FALSE;
				AnimationFrame frame;
				{
					LPTSTR pContext2 = NULL;
					LPTSTR p = _tcstok_s(pCur, TEXT(":"), &pContext2);
					if (p)
					{
						LPTSTR pContext3 = NULL;
						LPTSTR t = _tcstok_s(p, TEXT("-"), &pContext3);
						if (t)
						{
							frame.nIndexMin = _tstoi(t);
							frame.nIndexMin = max(0, min(nTotalCount, frame.nIndexMin));
							t = _tcstok_s(NULL, TEXT("-"), &pContext3);
							if (t)
							{
								frame.nIndexMax = _tstoi(t);
							}
							else
							{
								frame.nIndexMax = frame.nIndexMin;
							}
							frame.nIndexMax = max(0, min(nTotalCount, frame.nIndexMax));
						}
						else
						{
							//  error
							continue;
						}
					}
					else
					{
						//  error
						continue;
					}

					p = _tcstok_s(NULL, TEXT(":"), &pContext2);
					if (p)
					{
						LPTSTR pContext3 = NULL;
						LPTSTR t = _tcstok_s(p, TEXT("-"), &pContext3);
						if (t)
						{
							frame.nFrameTimeMin = _tstoi(t);
							frame.nFrameTimeMin = max(0, frame.nFrameTimeMin);
							t = _tcstok_s(NULL, TEXT("-"), &pContext3);
							if (t)
							{
								frame.nFrameTimeMax = _tstoi(t);
								frame.nFrameTimeMax = max(0, frame.nFrameTimeMax);
								if (frame.nFrameTimeMin > frame.nFrameTimeMax)
								{
									int a = frame.nFrameTimeMax;
									frame.nFrameTimeMax = frame.nFrameTimeMin;
									frame.nFrameTimeMin = a;
								}
							}
							else
							{
								frame.nFrameTimeMax = frame.nFrameTimeMin;
							}
						}
						else
						{
							frame.nFrameTimeMin = 0;
							frame.nFrameTimeMax = 0;
						}
					}
					else
					{
						frame.nFrameTimeMin = 0;
						frame.nFrameTimeMax = 0;
					}
					{
						//  VC2008 SP1 RpCoO΍
						AnimationFrame tmp = frame;
						frameList.push_back(tmp);
					}
				}
			}
		}
		_freea(pReadBuf);
	}
	nCurrentIndex = 0;

	if (fIsUseCustomSequence)
	{
		if (frameList.size() <= 0)
		{
			Unload();
			return FALSE;
		}
	}
	else
	{
		if (nTotalCount <= 0)
		{
			Unload();
			return FALSE;
		}
	}
	hKillEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (hKillEvent == NULL || hKillEvent == INVALID_HANDLE_VALUE)
	{
		Unload();
		return FALSE;
	}
	unsigned int id;
	hThread = (HANDLE) _beginthreadex(NULL, 0, ThreadProc, this, 0, &id);
	if (hThread == NULL || hThread == INVALID_HANDLE_VALUE)
	{
		Unload();
		return FALSE;
	}

	stFieldName = pszFieldName;
	return TRUE;
}

void AnimationEntry::Unload(void)
{
	if (hThread)
	{
		SetEvent(hKillEvent);
		WaitForSingleObject(hThread, INFINITE);
		CloseHandle(hThread);
		hThread = NULL;
	}

	if (hKillEvent)
	{
		CloseHandle(hKillEvent);
		hKillEvent = NULL;
	}

	stFieldName.clear();
	if (bmp)
	{
		delete bmp;
		bmp = NULL;
	}
#ifdef _USE_DIRECT2D_INTERNAL_
	if (pDecoder)
	{
		pDecoder->Release();
		pDecoder = NULL;
	}
	if (pTarget)
	{
		pTarget->Release();
		pTarget = NULL;
	}
	if (pImagingFactory)
	{
		pImagingFactory->Release();
		pImagingFactory = NULL;
	}
#endif

	sizFrame.cx = 0;
	sizFrame.cy = 0;
	nXCount = 0;
	nYCount = 0;
	nTotalCount = 0;
	nCurrentIndex = 0;
	nInterval = 0;
	nLoopCount = 0;
	nCurrentLoopCount = 0;
	fIsFrameBufferInitialized = FALSE;
	fIsResetOnRedraw = FALSE;
}

unsigned int __stdcall AnimationEntry::ThreadProc(void *lpvParam)
{
	AnimationEntry * pthis = (AnimationEntry *) lpvParam;
	if (pthis == NULL)
	{
		return -1;
	}

	if (pthis->fIsUseCustomSequence)
	{
		return pthis->ThreadProcSubCustomSequence();
	}
	else
	{
		return pthis->ThreadProcSub();
	}
}

unsigned int AnimationEntry::ThreadProcSub(void)
{
	DWORD dwStartTime = timeGetTime();
	DWORD dwEndTime = nInterval * nTotalCount * nLoopCount;

	for ( ; ; )
	{
		DWORD dwCurTime = timeGetTime();
		if (nInterval > 0)
		{
			if (dwEndTime && dwCurTime - dwStartTime > dwEndTime)
			{
				nCurrentIndex = nTotalCount - 1;
				DrawFrame();
				return 0;
			}
			DWORD index = (dwCurTime - dwStartTime) / (DWORD) nInterval;
			nCurrentIndex = index % nTotalCount;
		}
		else
		{
			nCurrentIndex++;
			if (nCurrentIndex >= nTotalCount)
			{
				nCurrentLoopCount++;
				if (nLoopCount <= 0 || nCurrentLoopCount < nLoopCount)
				{
					nCurrentIndex = 0;
				}
				else
				{
					break;
				}
			}
		}
		if (! DrawFrame())
		{
			if (fIsResetOnRedraw && (! fIsFrameBufferInitialized))
			{
				nCurrentIndex = 0;
				dwStartTime = timeGetTime();
				dwEndTime = nInterval * nTotalCount * nLoopCount;
				DrawFrame(TRUE);
				fIsFrameBufferInitialized = TRUE;
			}
		}
		{
			int nWaitTime = dwCurTime + nInterval - timeGetTime();
			DWORD dwRet = WaitForSingleObject(this->hKillEvent, (nWaitTime >= 0) ? nWaitTime : 0);
			if (dwRet != WAIT_TIMEOUT)
			{
				break;
			}
		}
	}
	return 0;
}

unsigned int AnimationEntry::ThreadProcSubCustomSequence(void)
{
	int nCurrentSequence = 0;
	srand((unsigned int) __rdtsc());
	for ( ; ; )
	{
		DWORD dwCurTime = timeGetTime();
		if (frameList[nCurrentSequence].nIndexMin == frameList[nCurrentSequence].nIndexMax)
		{
			nCurrentIndex = frameList[nCurrentSequence].nIndexMin;
		}
		else
		{
			int max = frameList[nCurrentSequence].nIndexMax - frameList[nCurrentSequence].nIndexMin;
			nCurrentIndex = (rand() % max) + frameList[nCurrentSequence].nIndexMin;
		}
		if (! DrawFrame())
		{
			if (fIsResetOnRedraw && (! fIsFrameBufferInitialized))
			{
				nCurrentSequence = 0;
				if (frameList[nCurrentSequence].nIndexMin == frameList[nCurrentSequence].nIndexMax)
				{
					nCurrentIndex = frameList[nCurrentSequence].nIndexMin;
				}
				else
				{
					int max = frameList[nCurrentSequence].nIndexMax - frameList[nCurrentSequence].nIndexMin;
					nCurrentIndex = (rand() % max) + frameList[nCurrentSequence].nIndexMin;
				}
				DrawFrame(TRUE);
				fIsFrameBufferInitialized = TRUE;
			}
		}
		{
			int nInterval;
			if (frameList[nCurrentSequence].nFrameTimeMin == frameList[nCurrentSequence].nFrameTimeMax)
			{
				nInterval = frameList[nCurrentSequence].nFrameTimeMin;
			}
			else
			{
				int max = frameList[nCurrentSequence].nFrameTimeMax - frameList[nCurrentSequence].nFrameTimeMin;
				nInterval = (rand() % max) + frameList[nCurrentSequence].nFrameTimeMin;
			}
			int nWaitTime = dwCurTime + nInterval - timeGetTime();
			DWORD dwRet = WaitForSingleObject(this->hKillEvent, (nWaitTime >= 0) ? nWaitTime : 0);
			if (dwRet != WAIT_TIMEOUT)
			{
				break;
			}
		}
		nCurrentSequence++;
		if (nCurrentSequence >= (int) frameList.size())
		{
			nCurrentLoopCount++;
			if (nLoopCount <= 0 || nCurrentLoopCount < nLoopCount)
			{
				nCurrentSequence = 0;
			}
			else
			{
				break;
			}
		}
	}
	return 0;
}

BOOL AnimationEntry::DrawFrame(BOOL fForce)
{
	BOOL ret = FALSE;
	if (g_callbacks)
	{
		//  fbhbN̂߁Â悤ɕύX
		if (g_callbacks->WaitLockFace(hFace, 1, &(this->hKillEvent), INFINITE) != WAIT_OBJECT_0 + 1)
		{
			return ret;
		}

		if (g_callbacks->GetDrawMode() == GenericPlugin::DrawModeDirect2D)
		{
			GenericPlugin::DrawItemD2D Item;
			Item = g_callbacks->GetFaceItemD2D(hFace, stFieldName.c_str());
			if ((Item.fIsVisible || fForce) && Item.pRenderTarget)
			{
				DrawFrameSubD2D(hFace, Item);
				Item.pRenderTarget->Release();
				g_callbacks->InvalidateFaceRect(hFace, Item.rcView);
				g_callbacks->UpdateFace(hFace);
				fIsFrameBufferInitialized = FALSE;
				ret = TRUE;
			}
		}
		else
		{
			GenericPlugin::DrawItem Item;
			Item = g_callbacks->GetFaceItem(hFace, stFieldName.c_str());
			if ((Item.fIsVisible || fForce) && Item.FrameBuffer)
			{
				DrawFrameSub(Item);
				g_callbacks->InvalidateFaceRect(hFace, Item.rcView);
				g_callbacks->UpdateFace(hFace);
				fIsFrameBufferInitialized = FALSE;
				ret = TRUE;
			}
		}
		g_callbacks->UnlockFace(hFace);
	}

	return ret;
}

void AnimationEntry::DrawFrameSub(const GenericPlugin::DrawItem& Item)
{
	if (! Item.FrameBuffer)
	{
		return;
	}
	Gdiplus::Graphics gr(Item.FrameBuffer);

	gr.Clear(0);

	int DestWidth = Item.FrameBuffer->GetWidth();
	int DestHeight = Item.FrameBuffer->GetHeight();

	SIZE sizDestFrame = { sizFrame.cx * dwViewScale / 100, sizFrame.cy * dwViewScale / 100 };
	Gdiplus::Rect rcDest(0, 0, DestWidth, DestHeight);
	if (sizDestFrame.cx <= DestWidth && sizDestFrame.cy <= DestHeight)
	{
		rcDest.Width = sizDestFrame.cx;
		rcDest.X = (DestWidth - sizDestFrame.cx) / 2;
		rcDest.Height = sizDestFrame.cy;
		rcDest.Y = (DestHeight - sizDestFrame.cy) / 2;
	}
	else
	{
		int y = sizDestFrame.cy * DestWidth / sizDestFrame.cx;
		if (y > DestHeight)
		{
			rcDest.Width = DestHeight * sizDestFrame.cx / sizDestFrame.cy;
			rcDest.X = (DestWidth - rcDest.Width) / 2;
		}
		else
		{
			rcDest.Height = y;
			rcDest.Y = (DestHeight - rcDest.Height) / 2;
		}
	}
	gr.SetCompositingMode(Gdiplus::CompositingModeSourceCopy);
	gr.SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
	if (rcDest.Width != sizDestFrame.cx || rcDest.Height != sizDestFrame.cy)
	{
		gr.SetInterpolationMode(Gdiplus::InterpolationModeBilinear);
	}
	else
	{
		gr.SetInterpolationMode(Gdiplus::InterpolationModeNearestNeighbor);
	}
	Gdiplus::ImageAttributes attr;
	const Gdiplus::ColorMatrix cmx = {
		1.0f, 0.0f, 0.0f, 0.0f, 0.0f,   // Red
		0.0f, 1.0f, 0.0f, 0.0f, 0.0f,   // Green
		0.0f, 0.0f, 1.0f, 0.0f, 0.0f,   // Blue
		0.0f, 0.0f, 0.0f, bCurrentAlpha / 255.0f, 0.0f,   // Alpha
		0.0f, 0.0f, 0.0f, 0.0f, 1.0f    // must be 1
	};
	attr.SetColorMatrix(&cmx, Gdiplus::ColorMatrixFlagsDefault, Gdiplus::ColorAdjustTypeBitmap); // Set ColorMatrix
	if (fIsAnimationImage)
	{
		bmp->SelectActiveFrame(&guidFrameDimension, nCurrentIndex);
		if (bCurrentAlpha < 255)
		{
			if (fNeedColorConvert)
			{
				Gdiplus::Bitmap tmp(sizFrame.cx, sizFrame.cy, PixelFormat32bppARGB);
				{
					Gdiplus::Graphics gr2(&tmp);
					gr.DrawImage(bmp, Gdiplus::Rect(0, 0, sizFrame.cx, sizFrame.cy), 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
				}
				gr.DrawImage(&tmp, rcDest, 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, &attr, NULL, NULL);
			}
			else
			{
				gr.DrawImage(bmp, rcDest, 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, &attr, NULL, NULL);
			}
		}
		else
		{
			gr.DrawImage(bmp, rcDest, 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
		
		}
	}
	else
	{
		Gdiplus::Point pt(sizFrame.cx * (nCurrentIndex % nXCount), sizFrame.cy * (nCurrentIndex / nXCount));
		if (bCurrentAlpha < 255)
		{
			if (fNeedColorConvert)
			{
				Gdiplus::Bitmap tmp(sizFrame.cx, sizFrame.cy, PixelFormat32bppARGB);
				{
					Gdiplus::Graphics gr2(&tmp);
					gr.DrawImage(bmp, Gdiplus::Rect(0, 0, sizFrame.cx, sizFrame.cy), 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
				}
				gr.DrawImage(&tmp, rcDest, pt.X, pt.Y, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, &attr, NULL, NULL);
			}
			else
			{
				gr.DrawImage(bmp, rcDest, pt.X, pt.Y, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, &attr, NULL, NULL);
			}
		}
		else
		{
			gr.DrawImage(bmp, rcDest, pt.X, pt.Y, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
		}
	}
}

void AnimationEntry::DrawFrameSubD2D(HANDLE hFace, const GenericPlugin::DrawItemD2D& Item)
{
	if (! Item.pRenderTarget)
	{
		return;
	}
#ifdef _USE_DIRECT2D_INTERNAL_
	D2D1_RECT_F rcSrc;
	if (fIsAnimationImage)
	{
		ExtractFrame(hFace, nCurrentIndex);
		rcSrc.left = 0.0f;
		rcSrc.top = 0.0f;
		rcSrc.right = (FLOAT) sizFrame.cx;
		rcSrc.bottom = (FLOAT) sizFrame.cy;
	}
	else
	{
		rcSrc.left = (FLOAT) (sizFrame.cx * (nCurrentIndex % nXCount));
		rcSrc.top = (FLOAT) (sizFrame.cy * (nCurrentIndex / nXCount));
		rcSrc.right = rcSrc.left + sizFrame.cx;
		rcSrc.bottom = rcSrc.top + sizFrame.cy;
	}
	if (! pTarget)
	{
		return;
	}
	D2D1_SIZE_F sizDest = Item.pRenderTarget->GetSize();
	SIZE sizDestFrame = { sizFrame.cx * dwViewScale / 100, sizFrame.cy * dwViewScale / 100 };
	D2D1_RECT_F rcDest = D2D1::RectF(0.0f, 0.0f, sizDest.width, sizDest.height);
	D2D1_SIZE_F sizSrc = D2D1::SizeF((FLOAT) sizDestFrame.cx, (FLOAT) sizDestFrame.cy);
	if (sizSrc.width <= sizDest.width && sizSrc.height <= sizDest.height)
	{
		rcDest.left = (sizDest.width - sizSrc.width) / 2.0f;
		rcDest.right = rcDest.left + sizSrc.width;
		rcDest.top = (sizDest.height - sizSrc.height) / 2.0f;
		rcDest.bottom = rcDest.top + sizSrc.height;
	}
	else
	{
		FLOAT y = sizSrc.height * sizDest.width / sizSrc.width;
		if (y > sizDest.height)
		{
			FLOAT x = sizDest.height * sizSrc.width / sizSrc.height;
			rcDest.left = (sizDest.width - x) / 2.0f;
			rcDest.right = rcDest.left + x;
		}
		else
		{
			rcDest.left = (sizDest.height - y) / 2.0f;
			rcDest.bottom = rcDest.left + y;
		}
	}
	D2D1_BITMAP_INTERPOLATION_MODE interMode = _SCALING_INTERPOLATION_MODE_;
#ifdef _USE_SCALING_LINEAR_
	if (dwViewScale > 100)
	{
		//  oCjAԂŊg傷ƁAlӂPhbgɃC邱ƂH
		//  N {ɂ邲Ƃɗ̈ 0.5 x N 傫Ɖ͗lBp邩ȂB
		//  jAXglCo[ɂΉ̂Aǂ炪ǂcc{̑ƍ킹Kv
		//  GPU ̐ݒɂ̂ȂAGTX 1080  Haswell ǂłԂB
		FLOAT diff = 0.5f * (dwScale / 100);
		rcDest.left -= diff;
		rcDest.top -= diff;
		rcDest.right += diff;
		rcDest.bottom += diff;
	}
#else
	if ((rcDest.right - rcDest.left) != sizDestFrame.cx || (rcDest.bottom - rcDest.top) != sizDestFrame.cy)
	{
		interMode = D2D1_BITMAP_INTERPOLATION_MODE_LINEAR;
	}
#endif
	ID2D1Bitmap *pBitmap = NULL;
	pTarget->GetBitmap(&pBitmap);
	Item.pRenderTarget->BeginDraw();
	Item.pRenderTarget->Clear();
	Item.pRenderTarget->DrawBitmap(pBitmap, &rcDest, bCurrentAlpha / 255.0f, interMode, &rcSrc);
	Item.pRenderTarget->EndDraw();
	pBitmap->Release();
#else
	Item.pRenderTarget->BeginDraw();
	ID2D1GdiInteropRenderTarget *pGdiRenderTarget = NULL;
	Item.pRenderTarget->QueryInterface(__uuidof(ID2D1GdiInteropRenderTarget), (LPVOID *) &pGdiRenderTarget);
	if (pGdiRenderTarget)
	{
		HDC hdc = NULL;
		pGdiRenderTarget->GetDC(D2D1_DC_INITIALIZE_MODE_CLEAR, &hdc);
		if (hdc)
		{
			D2D1_SIZE_F size = Item.pRenderTarget->GetSize();
			int DestWidth = (INT) size.width;
			int DestHeight = (INT) size.height;
			Gdiplus::Rect rcDest(0, 0, DestWidth, DestHeight);
			{
				Gdiplus::Graphics gr(hdc);
				if (sizFrame.cx <= DestWidth && sizFrame.cy <= DestHeight)
				{
					rcDest.Width = sizFrame.cx;
					rcDest.X = (DestWidth - sizFrame.cx) / 2;
					rcDest.Height = sizFrame.cy;
					rcDest.Y = (DestHeight - sizFrame.cy) / 2;
				}
				else
				{
					int y = sizFrame.cy * DestWidth / sizFrame.cx;
					if (y > DestHeight)
					{
						rcDest.Width = DestHeight * sizFrame.cx / sizFrame.cy;
						rcDest.X = (DestWidth - rcDest.Width) / 2;
					}
					else
					{
						rcDest.Height = y;
						rcDest.Y = (DestHeight - rcDest.Height) / 2;
					}
				}
				gr.SetCompositingMode(Gdiplus::CompositingModeSourceOver);
				gr.SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
				gr.SetInterpolationMode(Gdiplus::InterpolationModeBilinear);
				if (fIsAnimationImage)
				{
					bmp->SelectActiveFrame(&guidFrameDimension, nCurrentIndex);
					gr.DrawImage(bmp, rcDest, 0, 0, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
				}
				else
				{
					Gdiplus::Point pt(sizFrame.cx * (nCurrentIndex % nXCount), sizFrame.cy * (nCurrentIndex / nXCount));
					gr.DrawImage(bmp, rcDest, pt.X, pt.Y, sizFrame.cx, sizFrame.cy, Gdiplus::UnitPixel, NULL, NULL, NULL);
				}
			}
			{
				RECT rc = { rcDest.GetLeft(), rcDest.GetTop(), rcDest.GetRight(), rcDest.GetBottom() };
				pGdiRenderTarget->ReleaseDC(&rc);
			}
		}
		pGdiRenderTarget->Release();
	}
	Item.pRenderTarget->EndDraw();
#endif
}

bool AnimationEntry::SetMouseEvent(GenericPlugin::MouseEvent mouseEvent)
{
	BYTE bPrevAlpha = bCurrentAlpha;
	switch (mouseEvent)
	{
	case GenericPlugin::MouseEventNormal:
		{
			bCurrentAlpha = bAlphaNormal;
		}
		break;
	case GenericPlugin::MouseEventPushed:
		{
			bCurrentAlpha = bAlphaPushed;
		}
		break;
	case GenericPlugin::MouseEventOver:
		{
			bCurrentAlpha = bAlphaOver;
		}
		break;
	}
	return (bPrevAlpha != bCurrentAlpha);
}

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	switch (uMessage)
	{
	case WM_INITDIALOG:
		{
			AnimationData *data = (AnimationData *) lParam;
			SetProp(hwnd, PROP_ANIMATIONDATA, (HANDLE) data);
			g_callbacks->RegisterDialog(hwnd);
			ShowWindow(hwnd, SW_SHOWNORMAL);
		}
		break;
	case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
			case IDOK:
			case IDCANCEL:
				{
					DestroyWindow(hwnd);
				}
				break;
			}
		}
		break;
	case WM_DESTROY:
		{
			AnimationData *data = (AnimationData *) GetProp(hwnd, PROP_ANIMATIONDATA);
			g_callbacks->UnregisterDialog(hwnd);
			if (data)
			{
				data->hwndProperty = NULL;
			}
		}
		break;
	}

	return FALSE;
}

BOOL WINAPI FaceItemNotifyCallback(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFieldName, GenericPlugin::FaceItemNotify NotifyCode, LPCVOID pNotifyStruct)
{
	AnimationData *data = (AnimationData *) lpvUserData;

	if (! data)
	{
		return FALSE;
	}

	AnimationFaceList::iterator itList;
	AnimationList::iterator it;

	itList = data->FaceList.find(hFace);
	if (itList != data->FaceList.end())
	{
		it = itList->second.find(pszFieldName);
		if (it != itList->second.end())
		{
			if (NotifyCode == GenericPlugin::FaceItemNotifySizeChanged)
			{
				const GenericPlugin::FaceItemNotifySizeChangedStruct * p = (const GenericPlugin::FaceItemNotifySizeChangedStruct *) pNotifyStruct;
				if (p->mode == GenericPlugin::DrawModeDirect2D)
				{
					it->second->DrawFrameSubD2D(hFace, p->ItemD2D);
					if (p->ItemD2D.pRenderTarget)
					{
						p->ItemD2D.pRenderTarget->Release();
					}
				}
				else
				{
					it->second->DrawFrameSub(p->Item);
				}
			}
			else if (NotifyCode == GenericPlugin::FaceItemNotifyMouseEvent)
			{
				const GenericPlugin::FaceItemNotifyMouseEventStruct * p = (const GenericPlugin::FaceItemNotifyMouseEventStruct *) pNotifyStruct;
				if (it->second->SetMouseEvent(p->mouseEvent))
				{
					if (p->mode == GenericPlugin::DrawModeDirect2D)
					{
						it->second->DrawFrameSubD2D(hFace, p->ItemD2D);
						if (p->ItemD2D.pRenderTarget)
						{
							p->ItemD2D.pRenderTarget->Release();
						}
					}
					else
					{
						it->second->DrawFrameSub(p->Item);
					}
				}
			}
			return TRUE;
		}
	}

	return FALSE;
}

#ifdef _USE_DIRECT2D_INTERNAL_

BOOL AnimationEntry::CreateRenderTarget(HANDLE hFace, UINT32 uWidth, UINT32 uHeight)
{
	HRESULT hres = E_FAIL;

	ID2D1Bitmap * pBitmap = NULL;
	ID2D1RenderTarget *pRenderTarget = g_callbacks->GetRootRenderTarget(hFace);

	if (pRenderTarget)
	{
		const D2D1_SIZE_F sizef = D2D1::SizeF((FLOAT) uWidth, (FLOAT) uHeight);
		const D2D1_SIZE_U sizeu = D2D1::SizeU(uWidth, uHeight);
		const D2D1_PIXEL_FORMAT pf = D2D1::PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, D2D1_ALPHA_MODE_PREMULTIPLIED);
		hres = pRenderTarget->CreateCompatibleRenderTarget(&sizef, &sizeu, &pf, D2D1_COMPATIBLE_RENDER_TARGET_OPTIONS_GDI_COMPATIBLE, &pTarget);
		if (SUCCEEDED(hres) && pTarget)
		{
			return TRUE;
		}
	}

	return FALSE;
}

void AnimationEntry::ExtractFrame(HANDLE hFace, UINT uIndex)
{
	HRESULT hres;
	if (pDecoder)
	{
		int i;

		if (nCurrentFrameInBuffer != (int) uIndex)
		{
			for (i = nCurrentFrameInBuffer; i != (int) uIndex; )
			{
				i = (i + 1) % nTotalCount;
				IWICBitmapFrameDecode * pFrame = NULL;
				hres = pDecoder->GetFrame(i, &pFrame);
				if (pFrame)
				{
					IWICFormatConverter * pConverter = NULL;
					hres = pImagingFactory->CreateFormatConverter(&pConverter);
					if (pConverter)
					{
						hres = pConverter->Initialize(pFrame, GUID_WICPixelFormat32bppPBGRA, WICBitmapDitherTypeNone, NULL, 0.0f, WICBitmapPaletteTypeMedianCut);
						if (SUCCEEDED(hres))
						{
							ID2D1Bitmap * pBitmap = NULL;
							ID2D1RenderTarget *pRenderTarget = g_callbacks->GetRootRenderTarget(hFace);

							if (pRenderTarget)
							{
								hres = pRenderTarget->CreateBitmapFromWicBitmap(pConverter, NULL, &pBitmap);
								pRenderTarget->Release();
								D2D1_SIZE_F siz = pBitmap->GetSize();
								if (i == 0)
								{
									ID2D1Bitmap *pDest = NULL;
									pTarget->GetBitmap(&pDest);
									if (pDest)
									{
										pDest->CopyFromBitmap(NULL, pBitmap, NULL);
									}
								}
								else
								{
									PROPVARIANT prop;
									PropVariantInit(&prop);

									D2D1_RECT_F rect = D2D1::RectF(0.0, 0.0, (FLOAT) sizFrame.cx, (FLOAT) sizFrame.cy);

									IWICMetadataQueryReader *pFrameMetadataQueryReader = NULL;
									pFrame->GetMetadataQueryReader(&pFrameMetadataQueryReader);
									if (pFrameMetadataQueryReader)
									{
										hres = pFrameMetadataQueryReader->GetMetadataByName(L"/imgdesc/Left", &prop);
										if (SUCCEEDED(hres) && prop.vt == VT_UI2)
										{
											rect.left = static_cast<FLOAT>(prop.uiVal);
											PropVariantClear(&prop);
										}

										hres = pFrameMetadataQueryReader->GetMetadataByName(L"/imgdesc/Top", &prop);
										if (SUCCEEDED(hres) && prop.vt == VT_UI2)
										{
											rect.top = static_cast<FLOAT>(prop.uiVal);
											PropVariantClear(&prop);
										}
										hres = pFrameMetadataQueryReader->GetMetadataByName(L"/imgdesc/Width", &prop);
										if (SUCCEEDED(hres) && prop.vt == VT_UI2)
										{
											rect.right = static_cast<FLOAT>(prop.uiVal) + rect.left;
											PropVariantClear(&prop);
										}

										hres = pFrameMetadataQueryReader->GetMetadataByName(L"/imgdesc/Height", &prop);
										if (SUCCEEDED(hres) && prop.vt == VT_UI2)
										{
											rect.bottom = static_cast<FLOAT>(prop.uiVal) + rect.top;
											PropVariantClear(&prop);
										}
										pFrameMetadataQueryReader->Release();
									}

									pTarget->BeginDraw();
									pTarget->DrawBitmap(pBitmap, &rect, 1.0f, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, NULL);
									pTarget->EndDraw();
								}
								pBitmap->Release();
							}
						}
						pConverter->Release();
					}
					pFrame->Release();
				}
			}
			nCurrentFrameInBuffer = i;
		}
	}
}
#endif
