#include "..\\uLilith\\Plugin\\GenericPlugin.h"
#include <shlwapi.h>
#include "Resource.h"
#include <process.h>

#include <string>
#include <xstring>
#include <map>


#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "gdiplus.lib")
#pragma comment(lib, "winmm.lib")

#ifdef __INTEL_COMPILER
#pragma warning(disable:1028)
#pragma warning(disable:1738)
#pragma warning(disable:1744)
#endif

#define STR_LEN 1024
#define PROP_EFFECTDATA TEXT("Prop::EffectData")

#define LARGE_BUFFER_LEN 65536

namespace std
{
	class string_less_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class wstring_less_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) < 0);
		}
	};
	class string_greater_i
	{
	public:
		bool operator() (const std::string& x, const std::string& y) const
		{
			return (_stricmp(x.c_str(), y.c_str()) > 0);
		}
	};
	class wstring_greater_i
	{
	public:
		bool operator() (const std::wstring& x, const std::wstring& y) const
		{
			return (_wcsicmp(x.c_str(), y.c_str()) > 0);
		}
	};
};
#ifdef _UNICODE
#define _tstring std::wstring
#define _tstring_less_i std::wstring_less_i
#else
#define _tstring std::string
#define _tstring_less_i std::string_less_i
#endif


inline BOOL NVStringToBool(LPCTSTR str)
{
	if (_tcsicmp(str, TEXT("TRUE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("ENABLE")) == 0)
	{
		return TRUE;
	}
	else if (_tcsicmp(str, TEXT("On")) == 0)
	{
		return TRUE;
	}
	else if (_tstoi(str) == 1)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
inline BOOL GetPrivateProfileBool(LPCTSTR pszField, LPCTSTR pszKey, BOOL fDefault, LPCTSTR pszFileName)
{
	TCHAR szBuf[256];

	GetPrivateProfileString(pszField, pszKey, fDefault ? TEXT("TRUE") : TEXT("FALSE"), szBuf, 256, pszFileName);
	return NVStringToBool(szBuf);
}

enum EffectType
{
	EffectClear = 0,
	EffectMosaic,
	EffectMax,
};

struct EffectEntry
{
	EffectType type;
	int nMosaicSize;
};
typedef std::map<_tstring, EffectEntry, _tstring_less_i> EffectList;

typedef std::map<_tstring, EffectList> FaceEffectList;
struct EffectData
{
	FaceEffectList faceList;
	HWND hwndProperty;
};


extern "C"
{
	void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info);
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks);
void WINAPI UninitializePlugin(LPVOID lpvUserData);
HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner);
int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName);
void WINAPI FaceEffectCallback(LPVOID lpvUserData, HANDLE hFace, const GenericPlugin::FaceEffectCallbackStruct& Canvas, LPCTSTR pszFieldName);

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam);

static void ProcessClear(const GenericPlugin::DrawItem& Canvas);
static void ProcessClearD2D(const GenericPlugin::DrawItemD2D& Canvas);
static void ProcessMosaic(const GenericPlugin::DrawItem& Canvas, int nMosaicSize);
static void ProcessMosaicD2D(const GenericPlugin::DrawItemD2D& Canvas, int nMosaicSize);

GenericPlugin::GenericPluginInfo g_Info =
{
	0x200,
	TEXT("SampleFaceEffect"),
	TEXT("Northern Verse (Project9k)"),
	TEXT("tFCXGtFNgvOC̃Tvł"),
	InitializePlugin,
	UninitializePlugin,
	ShowProperty,
	OnFaceLoad,
	OnFaceUnload,
	NULL,//OnSoundOpen,
	NULL,//OnSoundChange,
	NULL,//OnSoundClose,
	NULL,//OnUpdateVisual,
	NULL,//OnSoundStatus,
	FaceEffectCallback,
};

GenericPlugin::PluginCallbacks *g_callbacks = NULL;
HINSTANCE g_hThisInstance = NULL;

BOOL APIENTRY DllMain(HINSTANCE hThisInstance, DWORD dwNotification, LPVOID lpReserved)
{
	if (dwNotification == DLL_PROCESS_ATTACH)
	{
		g_hThisInstance = hThisInstance;
		DisableThreadLibraryCalls(hThisInstance);
	}
	else if (dwNotification == DLL_PROCESS_DETACH)
	{
	}

	return TRUE;
}



void WINAPI GetPluginInfo(GenericPlugin::GenericPluginInfo& info)
{
	info = g_Info;
}

LPVOID WINAPI InitializePlugin(HWND hwndMain, const GenericPlugin::PluginCallbacks* callbacks)
{
	EffectData *data = new EffectData;
	data->faceList.clear();
	data->hwndProperty = NULL;

	g_callbacks = (GenericPlugin::PluginCallbacks *) callbacks;

	return data;
}

void WINAPI UninitializePlugin(LPVOID lpvUserData)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		data->faceList.clear();
		delete data;
	}
}

HWND WINAPI ShowProperty(LPVOID lpvUserData, HWND hwndOwner)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		if (data->hwndProperty)
		{
			SetForegroundWindow(data->hwndProperty);
			return data->hwndProperty;
		}
		else
		{
			data->hwndProperty = CreateDialogParam(g_hThisInstance, MAKEINTRESOURCE(IDD_PROPERTY), hwndOwner, PropertyDialogProc, (LPARAM) data);
			return data->hwndProperty;
		}
	}
	return NULL;
}

int WINAPI OnFaceLoad(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		FaceEffectList::iterator it;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		it = data->faceList.find(pszFaceName);
		if (it == data->faceList.end())
		{
			data->faceList.insert(make_pair(pszFaceName, EffectList()));
			it = data->faceList.find(pszFaceName);
		}
		if (it != data->faceList.end())
		{
			TCHAR szFieldName[STR_LEN];
			HANDLE hffi = g_callbacks->FindFirstFaceItem(hFace, TEXT("SampleMosaic"), szFieldName, STR_LEN);
			if (hffi)
			{
				do
				{
					EffectEntry mosaic;
					mosaic.type = EffectMosaic;
					mosaic.nMosaicSize = GetPrivateProfileInt(szFieldName, TEXT("MosaicSize"), 0, pszFileName);
					it->second.insert(std::make_pair(szFieldName, mosaic));
				}
				while (g_callbacks->FindNextFaceItem(hffi, szFieldName, STR_LEN));
				g_callbacks->FindFaceItemClose(hffi);
			}
			hffi = g_callbacks->FindFirstFaceItem(hFace, TEXT("SampleClear"), szFieldName, STR_LEN);
			if (hffi)
			{
				do
				{
					EffectEntry clear;
					clear.type = EffectClear;
					it->second.insert(std::make_pair(szFieldName, clear));
				}
				while (g_callbacks->FindNextFaceItem(hffi, szFieldName, STR_LEN));
				g_callbacks->FindFaceItemClose(hffi);
			}
		}
	}

	return 0;
}

int WINAPI OnFaceUnload(LPVOID lpvUserData, HANDLE hFace, LPCTSTR pszFileName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		FaceEffectList::iterator it;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		it = data->faceList.find(pszFaceName);
		if (it != data->faceList.end())
		{
			data->faceList.erase(it);
		}
	}

	return 0;
}

static INT_PTR CALLBACK PropertyDialogProc(HWND hwnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	switch (uMessage)
	{
	case WM_INITDIALOG:
		{
			EffectData *data = (EffectData *) lParam;
			SetProp(hwnd, PROP_EFFECTDATA, (HANDLE) data);
			g_callbacks->RegisterDialog(hwnd);
			ShowWindow(hwnd, SW_SHOWNORMAL);
		}
		break;
	case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
			case IDOK:
			case IDCANCEL:
				{
					DestroyWindow(hwnd);
				}
				break;
			}
		}
		break;
	case WM_DESTROY:
		{
			EffectData *data = (EffectData *) GetProp(hwnd, PROP_EFFECTDATA);
			g_callbacks->UnregisterDialog(hwnd);
			if (data)
			{
				data->hwndProperty = NULL;
			}
		}
		break;
	}

	return FALSE;
}


void WINAPI FaceEffectCallback(LPVOID lpvUserData, HANDLE hFace, const GenericPlugin::FaceEffectCallbackStruct& Canvas, LPCTSTR pszFieldName)
{
	EffectData *data = (EffectData *) lpvUserData;

	if (data)
	{
		EffectList::iterator it;
		FaceEffectList::iterator itList;
		LPCTSTR pszFaceName = g_callbacks->GetWindowNameFromHandle(hFace);

		itList = data->faceList.find(pszFaceName);
		if (itList != data->faceList.end())
		{
			it = itList->second.find(pszFieldName);
			if (it != itList->second.end())
			{
				if (it->second.type == EffectClear)
				{
					if (Canvas.mode == GenericPlugin::DrawModeGdiplus)
					{
						ProcessClear(Canvas.Item);
					}
					else if (Canvas.mode == GenericPlugin::DrawModeDirect2D)
					{
						ProcessClearD2D(Canvas.ItemD2D);
					}
				}
				else if (it->second.type == EffectMosaic)
				{
					int nMosaicSize = it->second.nMosaicSize;
					if (nMosaicSize > 0)
					{
						if (Canvas.mode == GenericPlugin::DrawModeGdiplus)
						{
							ProcessMosaic(Canvas.Item, nMosaicSize);
						}
						else if (Canvas.mode == GenericPlugin::DrawModeDirect2D)
						{
							ProcessMosaicD2D(Canvas.ItemD2D, nMosaicSize);
						}
					}
				}
			}
		}
	}
}

static void ProcessClear(const GenericPlugin::DrawItem& Canvas)
{
	Gdiplus::BitmapData bits;
	Gdiplus::Rect rc(Canvas.rcView.left, Canvas.rcView.top, Canvas.rcView.right - Canvas.rcView.left, Canvas.rcView.bottom - Canvas.rcView.top);

	Gdiplus::Status ret = Canvas.FrameBuffer->LockBits(&rc, Gdiplus::ImageLockModeRead | Gdiplus::ImageLockModeWrite, PixelFormat32bppARGB, &bits);
	if (ret == Gdiplus::Ok)
	{
		if (bits.PixelFormat == PixelFormat32bppARGB)
		{
			UINT i;

			for (i = 0; i < bits.Height; i++)
			{
				ZeroMemory(((LPBYTE) bits.Scan0) + i * bits.Stride, bits.Width * (32 / 8));
			}
		}

		Canvas.FrameBuffer->UnlockBits(&bits);
	}
}

static void ProcessClearD2D(const GenericPlugin::DrawItemD2D& Canvas)
{
	D2D1_COLOR_F color = D2D1::ColorF(0.0f, 0.0f, 0.0f, 0.0f);
	Canvas.pRenderTarget->BeginDraw();
	Canvas.pRenderTarget->PushAxisAlignedClip(D2D1::RectF((FLOAT) Canvas.rcView.left, (FLOAT) Canvas.rcView.top, (FLOAT) Canvas.rcView.right, (FLOAT) Canvas.rcView.bottom), D2D1_ANTIALIAS_MODE_ALIASED);
	Canvas.pRenderTarget->Clear(&color);
	Canvas.pRenderTarget->PopAxisAlignedClip();
	Canvas.pRenderTarget->EndDraw();
}

static void ProcessMosaic(const GenericPlugin::DrawItem& Canvas, int nMosaicSize)
{
	Gdiplus::BitmapData bits;
	Gdiplus::Rect rc(Canvas.rcView.left, Canvas.rcView.top, Canvas.rcView.right - Canvas.rcView.left, Canvas.rcView.bottom - Canvas.rcView.top);

	Gdiplus::Status ret = Canvas.FrameBuffer->LockBits(&rc, Gdiplus::ImageLockModeRead | Gdiplus::ImageLockModeWrite, PixelFormat32bppARGB, &bits);
	if (ret == Gdiplus::Ok)
	{
		if (bits.PixelFormat == PixelFormat32bppARGB)
		{
			UINT i, j;
			int m, n;
			UINT iMax = bits.Height / nMosaicSize;
			UINT jMax = bits.Width / nMosaicSize;

			for (i = 0; i < iMax; i++)
			{
				for (j = 0; j < jMax; j++)
				{
					UINT r = 0, g = 0, b = 0;

					for (m = 0; m < nMosaicSize; m++)
					{
						LPBYTE pCur = ((LPBYTE) bits.Scan0) + (i * nMosaicSize + m) * bits.Stride + j * nMosaicSize * (32 / 8);

						for (n = 0; n < nMosaicSize; n++)
						{
							r += pCur[n * (32 / 8) + 0];
							g += pCur[n * (32 / 8) + 1];
							b += pCur[n * (32 / 8) + 2];
						}
					}
					{
						int div = nMosaicSize * nMosaicSize;

						r /= div;
						g /= div;
						b /= div;
					}

					for (m = 0; m < nMosaicSize; m++)
					{
						LPBYTE pCur = ((LPBYTE) bits.Scan0) + (i * nMosaicSize + m) * bits.Stride + j * nMosaicSize * (32 / 8);

						for (n = 0; n < nMosaicSize; n++)
						{
							pCur[n * (32 / 8) + 0] = r;
							pCur[n * (32 / 8) + 1] = g;
							pCur[n * (32 / 8) + 2] = b;
						}
					}
				}
			}
		}
		Canvas.FrameBuffer->UnlockBits(&bits);
	}
}

static void ProcessMosaicD2D(const GenericPlugin::DrawItemD2D& Canvas, int nMosaicSize)
{
	int nWidth = Canvas.rcView.right - Canvas.rcView.left;
	int nHeight = Canvas.rcView.bottom - Canvas.rcView.top;
	int nPitch = nWidth * 4;

	ID2D1GdiInteropRenderTarget *pGdiRenderTarget = NULL;
	Canvas.pRenderTarget->QueryInterface(__uuidof(ID2D1GdiInteropRenderTarget), (LPVOID *) &pGdiRenderTarget);
	if (pGdiRenderTarget)
	{
		HDC hdc = NULL;
		Canvas.pRenderTarget->BeginDraw();
		HRESULT hres = pGdiRenderTarget->GetDC(D2D1_DC_INITIALIZE_MODE_COPY, &hdc);
		if (hdc)
		{
			BITMAPV4HEADER bmiHeader;

			ZeroMemory(&bmiHeader, sizeof(BITMAPV4HEADER));
			bmiHeader.bV4BitCount = 32;
			bmiHeader.bV4Height = -nHeight;
			bmiHeader.bV4Width = nWidth;
			bmiHeader.bV4Planes = 1;
			bmiHeader.bV4Size = sizeof(BITMAPV4HEADER);
			bmiHeader.bV4V4Compression = BI_BITFIELDS;
			bmiHeader.bV4RedMask = 0x00FF0000;
			bmiHeader.bV4GreenMask = 0x0000FF00;
			bmiHeader.bV4BlueMask = 0x000000FF;
			bmiHeader.bV4AlphaMask = 0xFF000000;
			LPVOID pvBits = NULL;
			HDC hdcBmp = CreateCompatibleDC(NULL);
			HBITMAP hbmp = CreateDIBSection(hdcBmp, (LPBITMAPINFO) &bmiHeader, DIB_RGB_COLORS, &pvBits, NULL, 0);
			int nSave = SaveDC(hdcBmp);
			SelectObject(hdcBmp, hbmp);

			BitBlt(hdcBmp, 0, 0, nWidth, nHeight, hdc, Canvas.rcView.left, Canvas.rcView.top, SRCCOPY);

			{
				UINT i, j;
				int m, n;
				UINT iMax = nHeight / nMosaicSize;
				UINT jMax = nWidth / nMosaicSize;

				for (i = 0; i < iMax; i++)
				{
					for (j = 0; j < jMax; j++)
					{
						UINT r = 0, g = 0, b = 0;

						for (m = 0; m < nMosaicSize; m++)
						{
							LPBYTE pCur = ((LPBYTE) pvBits) + (i * nMosaicSize + m) * nPitch + j * nMosaicSize * (32 / 8);

							for (n = 0; n < nMosaicSize; n++)
							{
								UINT a = pCur[n * (32 / 8) + 3];
								if (a > 0)
								{
									b += pCur[n * (32 / 8) + 0] * 255 / a;
									g += pCur[n * (32 / 8) + 1] * 255 / a;
									r += pCur[n * (32 / 8) + 2] * 255 / a;
								}
							}
						}
						{
							int div = nMosaicSize * nMosaicSize;

							r /= div;
							g /= div;
							b /= div;
						}

						for (m = 0; m < nMosaicSize; m++)
						{
							LPBYTE pCur = ((LPBYTE) pvBits) + (i * nMosaicSize + m) * nPitch + j * nMosaicSize * (32 / 8);

							for (n = 0; n < nMosaicSize; n++)
							{
								UINT a = pCur[n * (32 / 8) + 3];
								pCur[n * (32 / 8) + 0] = b * a / 255;
								pCur[n * (32 / 8) + 1] = g * a / 255;
								pCur[n * (32 / 8) + 2] = r * a / 255;
							}
						}
					}
				}
			}

			BitBlt(hdc, Canvas.rcView.left, Canvas.rcView.top, nWidth, nHeight, hdcBmp, 0, 0, SRCCOPY);

			RestoreDC(hdcBmp, nSave);
			DeleteObject(hbmp);
			DeleteDC(hdcBmp);

			pGdiRenderTarget->ReleaseDC(NULL);
		}
		Canvas.pRenderTarget->EndDraw();
		pGdiRenderTarget->Release();
	}
}
