//////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                              //
//  COM ݊ o̓Xg[ NX  \[Xt@C  for  NVFileLibrary                        //
//  IStream.cpp                                                                                 //
//                                                                                              //
//  Programed by Northern Verse (Project9k)                                                     //
//                                                                                              //
//////////////////////////////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////////////////////////////
//
//  wb_[t@C̃CN[h
//


#include "IStream.h"
#include <objidl.h>


Class_NVComStream::Class_NVComStream(void)
{
	invst = NULL;
	lRefCount = 1;
	return;
}

Class_NVComStream::~Class_NVComStream(void)
{
	if (invst)
	{
		invst->Release();
		invst = NULL;
	}
	return;
}


HRESULT Class_NVComStream::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject)
{
    if (ppvObject == NULL) {
        return E_POINTER;
    }

	if ( riid == IID_IStream)
    {
		*ppvObject = (IStream *) this;
    }
    else if (riid == IID_ISequentialStream)
	{
		*ppvObject = (ISequentialStream *) this;
	}
	else if (riid == IID_IUnknown)
	{
		*ppvObject = (IUnknown *) this;
	}
	else
	{
        return E_NOINTERFACE;
    }
	AddRef();

    return S_OK;
}

ULONG Class_NVComStream::AddRef(void)
{
	return InterlockedIncrement(&lRefCount);
}

ULONG Class_NVComStream::Release(void)
{
    LONG lRef = InterlockedDecrement(&lRefCount);
    if (lRef == 0)
	{
        delete this;
    }
    return lRef;
}

HRESULT Class_NVComStream::Read(void __RPC_FAR *pv, ULONG cb, ULONG __RPC_FAR *pcbRead)
{
	if (NVSafeLock(invst) == NV_OK)
	{
		NVRESULT nvres;
		nvres = invst->Read(pv, cb, pcbRead);
		NVSafeUnLock(invst);
		if (nvres && nvres != VSTMERR_VALUE_ENDOFDATA)
		{
			return E_INVALIDARG;
		}
		else
		{
			return S_OK;
		}
	}

	return E_INVALIDARG;
}

HRESULT Class_NVComStream::Write(const void __RPC_FAR *pv, ULONG cb, ULONG __RPC_FAR *pcbWritten)
{
	if (NVSafeLock(invst) == NV_OK)
	{
		NVRESULT nvres;
		nvres = invst->Write((void *)pv, cb, pcbWritten);
		NVSafeUnLock(invst);
		return (nvres) ? E_INVALIDARG : S_OK;
	}

	return E_INVALIDARG;
}

HRESULT Class_NVComStream::Seek(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER __RPC_FAR *plibNewPosition)
{
	DWORD dwFlag;
	switch (dwOrigin)
	{
	case STREAM_SEEK_CUR:
		dwFlag = NVSTREAM_SEEKCURRENT;
		break;
	case STREAM_SEEK_END:
		dwFlag = NVSTREAM_SEEKEND;
		break;
	default:
		dwFlag = NVSTREAM_SEEKBEGIN;
	}

	if (NVSafeLock(invst) == NV_OK)
	{
		NVRESULT nvres;

		nvres = invst->SetCurrentPosition(dlibMove.QuadPart, dwFlag);
		if (nvres)
		{
			NVSafeUnLock(invst);
			return E_INVALIDARG;
		}
		if (plibNewPosition && plibNewPosition != INVALID_HANDLE_VALUE)
		{
			nvres = invst->GetCurrentPosition((LPQLONG) &plibNewPosition->QuadPart, NVSTREAM_SEEKBEGIN);
			if (nvres)
			{
				NVSafeUnLock(invst);
				return E_INVALIDARG;
			}
		}
		NVSafeUnLock(invst);
		return S_OK;
	}

	return E_INVALIDARG;
}

HRESULT Class_NVComStream::SetSize(ULARGE_INTEGER libNewSize)
{
	if (NVSafeLock(invst) == NV_OK)
	{
		NVRESULT nvres;

		nvres = invst->SetDataSize(libNewSize.QuadPart);
		NVSafeUnLock(invst);
		return (nvres) ? E_INVALIDARG : S_OK;
	}

	return E_INVALIDARG;
}

HRESULT Class_NVComStream::CopyTo(IStream __RPC_FAR *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER __RPC_FAR *pcbRead, ULARGE_INTEGER __RPC_FAR *pcbWritten)
{

	if (NVSafeLock(invst) == NV_OK)
	{
		DWORD dwRead;
		QWORD qwReadTotal = 0;
		DWORD dwWrite;
		QWORD qwWriteTotal =0;
		BYTE lpBuf[65536];
		QWORD qwRemain = cb.QuadPart;

		while (qwRemain)
		{
			DWORD dwSize = (DWORD) min(65536, qwRemain);
			invst->Read(lpBuf, dwSize, &dwRead);
			pstm->Write(lpBuf, dwRead, &dwWrite);
			qwReadTotal += dwRead;
			qwWriteTotal += dwWrite;
			if ((dwRead != dwSize) || (dwWrite != dwSize))
			{
				break;
			}
			qwRemain -= dwWrite;
		}

		if (pcbRead)
		{
			pcbRead->QuadPart = qwReadTotal;
		}

		if (pcbWritten)
		{
			pcbWritten->QuadPart = qwWriteTotal;
		}

		NVSafeUnLock(invst);
		return (qwRemain) ? E_INVALIDARG : S_OK;
	}

	if (pcbRead)
	{
		pcbRead->QuadPart = 0;
	}

	if (pcbWritten)
	{
		pcbWritten->QuadPart = 0;
	}

	return E_INVALIDARG;
}
		
HRESULT Class_NVComStream::Commit(DWORD grfCommitFlags)
{
	return S_OK;
}
		
HRESULT Class_NVComStream::Revert( void)
{
	return E_INVALIDARG;
}

HRESULT Class_NVComStream::LockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
	return STG_E_INVALIDFUNCTION;
}

HRESULT Class_NVComStream::UnlockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType)
{
	return STG_E_INVALIDFUNCTION;
}

HRESULT Class_NVComStream::Stat(STATSTG __RPC_FAR *pstatstg, DWORD grfStatFlag)
{
	NVRESULT nvres;

	ZeroMemory(pstatstg, sizeof(STATSTG));
	if (grfStatFlag != STATFLAG_NONAME)
	{
		TCHAR lpszName[NV_PATHMAX];
		DWORD dwWideLen;
		LPWSTR lpwszWide;

		nvres = invst->GetDataName(lpszName, NV_PATHMAX);
		if (nvres)
		{
			return STG_E_ACCESSDENIED;
		}
#ifdef _UNICODE
		dwWideLen = (DWORD) _tcslen(lpszName) + 1;
		lpwszWide = (LPWSTR) CoTaskMemAlloc(dwWideLen * sizeof(WCHAR));
		_tcscpy_s(lpwszWide, dwWideLen, lpszName);
#else
		dwWideLen = MultiByteToWideChar(CP_ACP, 0, lpszName, -1, NULL, 0);
		lpwszWide = (LPWSTR) CoTaskMemAlloc(dwWideLen * sizeof(WCHAR));
		MultiByteToWideChar(CP_ACP, 0, lpszName, -1, lpwszWide, dwWideLen);
#endif
		pstatstg->pwcsName = lpwszWide;
	}
	else
	{
		pstatstg->pwcsName = NULL;
	}

	pstatstg->type = STGTY_STORAGE;
	//pstatstg->type = STGTY_STREAM;
	invst->GetDataSize(&pstatstg->cbSize.QuadPart);
	pstatstg->mtime.dwLowDateTime = 0;
	pstatstg->mtime.dwHighDateTime = 0;
	pstatstg->atime.dwLowDateTime = 0;
	pstatstg->atime.dwHighDateTime = 0;
	pstatstg->ctime.dwLowDateTime = 0;
	pstatstg->ctime.dwHighDateTime = 0;
	pstatstg->grfMode = 0;
	pstatstg->grfLocksSupported = 0;
	pstatstg->clsid = CLSID_NULL;
	pstatstg->grfStateBits = 0;
	pstatstg->reserved = 0;

	return S_OK;
}

HRESULT Class_NVComStream::Clone(IStream __RPC_FAR *__RPC_FAR *ppstm)
{
	if (! ppstm)
	{
		return E_POINTER;
	}

	if (NVSafeLock(invst) == NV_OK)
	{
		HRESULT hres;
		NVRESULT nvres;
		INVSTREAM invstClone;
		INVCOMSTREAM invcomst;

		nvres = invst->CreateAlias(&invstClone);
		NVSafeUnLock(invst);
		if (nvres)
		{
			*ppstm = NULL;
			return E_OUTOFMEMORY;
		}

		invcomst = new Class_NVComStream;
		hres = invcomst->CreateFromNVStream(invstClone);
		if (FAILED(hres))
		{
			invstClone->Release();
			*ppstm = NULL;
			invcomst->Release();
			return E_OUTOFMEMORY;
		}
		invstClone->Release();

		hres = invcomst->QueryInterface(IID_IStream, (LPVOID *)ppstm);
		if (FAILED(hres))
		{
			*ppstm = NULL;
			invcomst->Release();
			return E_OUTOFMEMORY;
		}

		invcomst->Release();
		return S_OK;
	}

	return E_INVALIDARG;
}

HRESULT Class_NVComStream::CreateFromNVStream(INVSTREAM invstSrc)
{
	if (invst)
	{
		invst->Release();
	}

	invstSrc->AddRef();
	invst = invstSrc;
	return S_OK;
}

