//////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                              //
//  NVWaveFileLibrary CW[ \[Xt@C  for  NVWaveFileLibrary                   //
//  NVWave.cpp                                                                                  //
//                                                                                              //
//  Programed by Northern Verse (Project9k)                                                     //
//                                                                                              //
//////////////////////////////////////////////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////////////////////////////////////////
//
//  wb_[t@C̃CN[h
//

#include "NVWaveDef.h"
#include "NVPCM.h"
#include "NVACM.h"
#include "NVPCMEnc.h"



#ifdef __cplusplus
extern "C" {
#endif
//  Generic
NVRESULT WINAPI VSSGetPluginInfo(LPVSSPLUGININFO lpInfo);

//  VSS Category D
NVRESULT WINAPI VSSDecGetInfo(LPVSSDECINFO lpInfo);
NVRESULT WINAPI VSSDecIsSupported(INVSTREAM invst, DWORD dwFourCC);
NVRESULT WINAPI VSSDecCreateInstance(DWORD dwFourCC, LPIVCOM lpivcom);
HWND WINAPI VSSDecConfigDialog(DWORD dwFourCC, HINSTANCE hInstance, HWND hwndOwner);

//  VSS Category E
NVRESULT WINAPI VSSEncGetInfo(LPVSSENCINFO lpInfo);
NVRESULT WINAPI VSSEncCreateInstance(DWORD dwFourCC, LPIVCOM lpivcom);
HWND WINAPI VSSEncConfigDialog(DWORD dwFourCC, HINSTANCE hInstance, HWND hwndOwner, LPVOID * lplpvParam);
NVRESULT WINAPI VSSEncFreeParam(LPVOID lpvParam);
NVRESULT WINAPI VSSEncGetParam(DWORD dwFourCC, LPVOID *lplpvParam);
NVRESULT WINAPI VSSEncCopyParam(DWORD dwFourCC, LPVOID *lplpvDest, LPVOID lpvSrc);

#ifdef __cplusplus
}
#endif


//////////////////////////////////////////////////////////////////////////////////////////////////
//
//  O[oϐ̒`
//

HINSTANCE hinstNVWave;					//  DLL W[̃CX^Xnh
LONG lNVWaveAttatchedProcess = 0;				//  A^b`ꂽvZX̐

#define SUPPORT_COUNT 2

const static DWORD lpdwSupportFourCC[SUPPORT_COUNT] = { NVPCM_FOURCC, NVACM_FOURCC };
const static LPCTSTR lpcszSupportFileType[SUPPORT_COUNT] =
{
	TEXT("RIFF PCM WaveFile"), TEXT("RIFF ACM WaveFile"),
};
const static VSSPLUGININFO vssInfo =
{
	0x3000, TEXT("NVWaveFileLibrary"), TEXT("Northern Verse (Project9k)"),
	TEXT("http://www.project9k.jp/"), TEXT("PCM/ACM WaveFile Reader/Decoder/Writer"),
	VSSDecGetInfo, VSSEncGetInfo, NULL,
};

const static VSSDECINFO vssDecInfo =
{
	SUPPORT_COUNT, lpdwSupportFourCC, lpcszSupportFileType,
	VSSDecIsSupported, VSSDecCreateInstance, VSSDecConfigDialog,
};

#define ENC_SUPPORT_COUNT 1

const static DWORD lpdwEncSupportedFourCC[ENC_SUPPORT_COUNT] = { NVPCMENC_FOURCC };
const static LPCTSTR lpcszEncSupportedFileType[ENC_SUPPORT_COUNT] = { TEXT("RIFF PCM WaveFile") };
const static LPCTSTR lpcszEncDefaultExtension[ENC_SUPPORT_COUNT] = { TEXT("wav") };
const static VSSENCINFO vssEncInfo =
{
	ENC_SUPPORT_COUNT, lpdwEncSupportedFourCC, lpcszEncSupportedFileType, lpcszEncDefaultExtension,
	VSSEncCreateInstance, VSSEncConfigDialog, VSSEncFreeParam, VSSEncGetParam, VSSEncCopyParam,
};


TCHAR szConfigFileName[NV_PATHMAX];

//////////////////////////////////////////////////////////////////////////////////////////////////
//
//  vOC֐̒`
//

NVRESULT WINAPI VSSGetPluginInfo(LPVSSPLUGININFO lpInfo)
{
	if (! lpInfo)
	{
		return VAFERR_INVALIDPARAM;
	}

	CopyMemory(lpInfo, &vssInfo, sizeof(VSSPLUGININFO));
	return NV_OK;
}

NVRESULT WINAPI VSSDecGetInfo(LPVSSDECINFO lpInfo)
{
	if (! lpInfo)
	{
		return VAFERR_INVALIDPARAM;
	}

	CopyMemory(lpInfo, &vssDecInfo, sizeof(VSSDECINFO));
	return NV_OK;
}

NVRESULT WINAPI VSSDecIsSupported(INVSTREAM invst, DWORD dwFourCC)
{
	NVRESULT nvres;
	LONG lType;
	QLONG qlStartOffset;
	QLONG qlDataSize;

	switch (dwFourCC)
	{
	case NVPCM_FOURCC:
		{
			nvres = CheckPCMFileType(invst, &lType, &qlStartOffset, &qlDataSize, NULL);
			if (nvres == NV_OK)
			{
				return NV_OK;
			}
		}
		break;
	case NVACM_FOURCC:
		{
			nvres = CheckACMFileType(invst, &lType, &qlStartOffset, &qlDataSize, NULL, NULL, NULL);
			if (nvres == NV_OK)
			{
				return NV_OK;
			}
		}
		break;
	}

	return VAFERR_UNSUPPORTED;
}

NVRESULT WINAPI VSSDecCreateInstance(DWORD dwFourCC, LPIVCOM lpivcom)
{
	switch (dwFourCC)
	{
	case NVPCM_FOURCC:
		{
			return CreateInstance_NVPCM(lpivcom);
		}
	case NVACM_FOURCC:
		{
			return CreateInstance_NVACM(lpivcom);
		}
	default:
		return VAFERR_UNSUPPORTED;
	}
}

HWND WINAPI VSSDecConfigDialog(DWORD dwFourCC, HINSTANCE hInstance, HWND hwndOwner)
{
	switch (dwFourCC)
	{
	case NVPCM_FOURCC:
		{
			MessageBox(hwndOwner, lpcszSupportFileType[0], vssInfo.lpcszLibraryName, MB_OK | MB_TOPMOST);
		}
		return NULL;
	case NVACM_FOURCC:
		{
			MessageBox(hwndOwner, lpcszSupportFileType[1], vssInfo.lpcszLibraryName, MB_OK | MB_TOPMOST);
		}
		return NULL;
	default:
		return NULL;
	}
}

NVRESULT WINAPI VSSEncGetInfo(LPVSSENCINFO lpInfo)
{
	if (! lpInfo)
	{
		return VAFERR_INVALIDPARAM;
	}

	CopyMemory(lpInfo, &vssEncInfo, sizeof(VSSENCINFO));
	return NV_OK;
}

NVRESULT WINAPI VSSEncCreateInstance(DWORD dwFourCC, LPIVCOM lpivcom)
{
	switch (dwFourCC)
	{
	case NVPCMENC_FOURCC:
		{
			return CreateInstance_NVPCMEnc(lpivcom);
		}
	default:
		return VAFERR_UNSUPPORTED;
	}
}

HWND WINAPI VSSEncConfigDialog(DWORD dwFourCC, HINSTANCE hInstance, HWND hwndOwner, LPVOID * lplpvParam)
{
	switch (dwFourCC)
	{
	case NVPCMENC_FOURCC:
		{
			return NVPCMEncConfigDialog(hInstance, hwndOwner, lplpvParam);
		}
	default:
		return NULL;
	}
}

NVRESULT WINAPI VSSEncFreeParam(LPVOID lpvParam)
{
	NVMemFree(lpvParam);
	return NV_OK;
}

NVRESULT WINAPI VSSEncGetParam(DWORD dwFourCC, LPVOID *lplpvParam)
{
	switch (dwFourCC)
	{
	case NVPCMENC_FOURCC:
		{
			*lplpvParam = NVPCMEncGetSetting();
			return (*lplpvParam) ? NV_OK : NVERR_UNDEFINED;
		}
	default:
		return NVERR_UNDEFINED;
	}
}

NVRESULT WINAPI VSSEncCopyParam(DWORD dwFourCC, LPVOID *lplpvDest, LPVOID lpvSrc)
{
	switch (dwFourCC)
	{
	case NVPCMENC_FOURCC:
		{
			*lplpvDest = NVPCMEncCopySetting(lpvSrc);
			return (*lplpvDest) ? NV_OK : NVERR_UNDEFINED;
		}
	default:
		return NVERR_UNDEFINED;
	}
}




//////////////////////////////////////////////////////////////////////////////////////////////////
//
//  O[o֐̒`
//


BOOL APIENTRY DllMain(HINSTANCE hThisInstance, DWORD dwNotification, LPVOID lpReserved)
{
	//  vZX̃A^b`^f^b`̏
	if (dwNotification == DLL_PROCESS_ATTACH)
	{
		//  A^b`ꂽvZX̃JEg
		if (InterlockedIncrement(&lNVWaveAttatchedProcess) == 1)
		{
			hinstNVWave = hThisInstance;		//  CX^Xnh̕ۑ

			NVGetConfigFileDirectory(szConfigFileName, NV_PATHMAX);
			PathAppend(szConfigFileName, TEXT("Encoder.cfg"));

			//  Xbh̃A^b`ʒmȂ悤ɂ
			DisableThreadLibraryCalls(hThisInstance);
		}
	}
	else if (dwNotification == DLL_PROCESS_DETACH)
	{
		//  A^b`ꂽvZX̃JEg
		//  ÓIf[^̉
		if (InterlockedDecrement(&lNVWaveAttatchedProcess) == 0)
		{
		}
	}

	return TRUE;
}

